# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Apply data augmentation on the Liver Tumor Segmentation (LiTS) dataset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import os

from absl import app
from absl import flags

from six.moves import range
import tensorflow.compat.v1 as tf  # tf

# pylint: disable=g-direct-tensorflow-import,g-direct-third-party-import
from mesh_tensorflow.experimental import data_aug_lib

FLAGS = flags.FLAGS

flags.DEFINE_string('input_file_pattern', '', 'Path to input CT scans.')
flags.DEFINE_string('output_folder', '', 'Path to output folder.')
flags.DEFINE_string('output_file_prefix',
                    'augmented', 'Filename prefix.')

flags.DEFINE_integer('ct_resolution', 128,
                     'Resolution of CT images along depth, height and '
                     'width dimensions.')

flags.DEFINE_integer('num_data_aug', 1000,
                     'The number of data augmentation output.')
flags.DEFINE_integer('process_no', None, 'Which process number I am.')

flags.DEFINE_float('gen_fake_probability', 0.50,
                   'How much to scale intensities of lesion/non-lesion areas.')
flags.DEFINE_float('min_fake_lesion_ratio', 0.05,
                   'Minimum amount of synthetic lession in liver.')


def _dataset_creator():
  """Returns an unbatched dataset."""
  def _parser_fn(serialized_example):
    """Parses a single tf.Example into image and label tensors."""
    features = {}
    features['image/ct_image'] = tf.FixedLenFeature([], tf.string)
    features['image/label'] = tf.FixedLenFeature([], tf.string)
    parsed = tf.parse_single_example(serialized_example, features=features)

    image = tf.decode_raw(parsed['image/ct_image'], tf.float32)
    label = tf.decode_raw(parsed['image/label'], tf.float32)

    # Preprocess color, clip to 0 ~ 1.
    image = tf.clip_by_value(image / 1024.0 + 0.5, 0, 1)

    spatial_dims = [FLAGS.ct_resolution] * 3
    image = tf.reshape(image, spatial_dims)
    label = tf.reshape(label, spatial_dims)

    image, label = data_aug_lib.maybe_gen_fake_data_based_on_real_data(
        image, label, FLAGS.ct_resolution,
        FLAGS.min_fake_lesion_ratio, FLAGS.gen_fake_probability)

    return image, label

  dataset = tf.data.Dataset.list_files(
      FLAGS.input_file_pattern, shuffle=True).repeat()
  dataset = dataset.apply(functools.partial(
      tf.data.TFRecordDataset, compression_type='GZIP'))
  dataset = dataset.shuffle(2).map(_parser_fn, num_parallel_calls=2)

  return dataset


def save_to_tfrecord(image, label, process_no, idx,
                     output_path, output_file_prefix):
  """Save to TFRecord."""
  d_feature = {}
  d_feature['image/ct_image'] = tf.train.Feature(
      bytes_list=tf.train.BytesList(value=[image.reshape([-1]).tobytes()]))
  d_feature['image/label'] = tf.train.Feature(
      bytes_list=tf.train.BytesList(value=[label.reshape([-1]).tobytes()]))

  example = tf.train.Example(features=tf.train.Features(feature=d_feature))
  serialized = example.SerializeToString()

  result_file = os.path.join(
      output_path,
      '{}-{}-{}.tfrecords'.format(output_file_prefix, process_no, idx))
  options = tf.python_io.TFRecordOptions(
      tf.python_io.TFRecordCompressionType.GZIP)
  with tf.python_io.TFRecordWriter(result_file, options=options) as w:
    w.write(serialized)


def apply_data_aug():
  """Apply data augmentation and save augmented results."""
  if not tf.gfile.IsDirectory(FLAGS.output_folder):
    tf.gfile.MakeDirs(FLAGS.output_folder)

  dataset = _dataset_creator()
  ds_iterator = dataset.make_initializable_iterator()
  image, label = ds_iterator.get_next()

  with tf.Session() as sess:
    sess.run(ds_iterator.initializer)
    for idx in range(FLAGS.num_data_aug):
      image_np, label_np = sess.run([image, label])
      save_to_tfrecord(
          image_np, label_np, FLAGS.process_no, idx,
          FLAGS.output_folder, FLAGS.output_file_prefix)
  return


def main(argv):
  del argv
  apply_data_aug()


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mesh_tensorflow.auto_mtf.graph_interface."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import mesh_tensorflow as mtf
from mesh_tensorflow.auto_mtf import graph_interface
import tensorflow.compat.v1 as tf
from tensorflow.core.framework import cost_graph_pb2
from tensorflow.core.framework import tensor_shape_pb2
from tensorflow.core.framework import types_pb2


class GraphInterfaceTest(tf.test.TestCase):

  def setUp(self):
    super(GraphInterfaceTest, self).setUp()
    self._cost_graph = cost_graph_pb2.CostGraphDef(
        node=[
            cost_graph_pb2.CostGraphDef.Node(
                name="X",
                device="/device:CPU:0",
                id=0,
                output_info=[
                    cost_graph_pb2.CostGraphDef.Node.OutputInfo(
                        size=48,
                        alias_input_port=-1,
                        dtype=types_pb2.DT_INT32,
                        shape=tensor_shape_pb2.TensorShapeProto(
                            dim=[
                                tensor_shape_pb2.TensorShapeProto.Dim(size=3),
                                tensor_shape_pb2.TensorShapeProto.Dim(size=4),
                            ]
                        )
                    ),
                ],
            ),
            cost_graph_pb2.CostGraphDef.Node(
                name="Y",
                device="/device:CPU:0",
                id=1,
                output_info=[
                    cost_graph_pb2.CostGraphDef.Node.OutputInfo(
                        size=80,
                        alias_input_port=-1,
                        dtype=types_pb2.DT_INT32,
                        shape=tensor_shape_pb2.TensorShapeProto(
                            dim=[
                                tensor_shape_pb2.TensorShapeProto.Dim(size=4),
                                tensor_shape_pb2.TensorShapeProto.Dim(size=5),
                            ]
                        )
                    ),
                ],
            ),
            cost_graph_pb2.CostGraphDef.Node(
                name="Z1",
                device="/device:CPU:0",
                id=2,
                input_info=[
                    cost_graph_pb2.CostGraphDef.Node.InputInfo(
                        preceding_node=0,
                        preceding_port=0,
                    ),
                    cost_graph_pb2.CostGraphDef.Node.InputInfo(
                        preceding_node=1,
                        preceding_port=0,
                    ),
                ],
                output_info=[
                    cost_graph_pb2.CostGraphDef.Node.OutputInfo(
                        size=60,
                        alias_input_port=-1,
                        dtype=types_pb2.DT_INT32,
                        shape=tensor_shape_pb2.TensorShapeProto(
                            dim=[
                                tensor_shape_pb2.TensorShapeProto.Dim(size=3),
                                tensor_shape_pb2.TensorShapeProto.Dim(size=5),
                            ]
                        )
                    ),
                ],
                is_final=True,
            ),
            cost_graph_pb2.CostGraphDef.Node(
                name="Z2",
                device="/device:CPU:0",
                id=3,
                input_info=[
                    cost_graph_pb2.CostGraphDef.Node.InputInfo(
                        preceding_node=0,
                        preceding_port=0,
                    ),
                    cost_graph_pb2.CostGraphDef.Node.InputInfo(
                        preceding_node=1,
                        preceding_port=0,
                    ),
                ],
                output_info=[
                    cost_graph_pb2.CostGraphDef.Node.OutputInfo(
                        size=60,
                        alias_input_port=-1,
                        dtype=types_pb2.DT_INT32,
                        shape=tensor_shape_pb2.TensorShapeProto(
                            dim=[
                                tensor_shape_pb2.TensorShapeProto.Dim(size=3),
                                tensor_shape_pb2.TensorShapeProto.Dim(size=5),
                            ]
                        )
                    ),
                ],
            ),
        ]
    )
    self._sizeless_cost_graph = self.StripCostGraphDef(
        self._cost_graph, "SIZES")
    self._deviceless_cost_graph = self.StripCostGraphDef(
        self._cost_graph, "DEVICES")

    self._cost_graph_string = self._cost_graph.SerializeToString()
    self._sizeless_cost_graph_string = (
        self._sizeless_cost_graph.SerializeToString())
    self._deviceless_cost_graph_string = (
        self._deviceless_cost_graph.SerializeToString())

  def StripCostGraphDef(self, cost_graph, to_strip):
    """Removes fields from a CostGraphDef protobuf.

    Helper method to reduce the initialization of CostGraphDef(s).

    Args:
      cost_graph: a CostGraphDef to strip.
      to_strip: a string, either "SIZES" or "DEVICES".

    Returns:
      a new CostGraphDef with either size information or device information
          stripped, as appropriate.
    """
    new_cost_graph = cost_graph_pb2.CostGraphDef()
    new_cost_graph.CopyFrom(cost_graph)
    for node in new_cost_graph.node:
      if to_strip == "SIZES":
        for output_info in node.output_info:
          output_info.size = 0
          output_info.ClearField("shape")
      if to_strip == "DEVICES":
        node.ClearField("device")
    return new_cost_graph

  def VerifyGraphInterface(self, graph):
    self.assertEqual(list(graph.get_all_operation_names()),
                     ["X", "Y", "Z1", "Z2"])

    self.assertEqual(list(graph.get_operation_input_names("X")), [])
    self.assertEqual(list(graph.get_operation_input_names("Y")), [])
    self.assertEqual(list(graph.get_operation_input_names("Z1")),
                     ["X:0", "Y:0"])
    self.assertEqual(list(graph.get_operation_input_names("Z2")),
                     ["X:0", "Y:0"])

    self.assertEqual(list(graph.get_operation_output_names("X")), ["X:0"])
    self.assertEqual(list(graph.get_operation_output_names("Y")), ["Y:0"])
    self.assertEqual(list(graph.get_operation_output_names("Z1")), ["Z1:0"])
    self.assertEqual(list(graph.get_operation_output_names("Z2")), ["Z2:0"])

    self.assertEqual(list(graph.get_all_tensor_names()),
                     ["X:0", "Y:0", "Z1:0", "Z2:0"])

    self.assertEqual(graph.get_tensor_dtype("X:0"), tf.int32)
    self.assertEqual(graph.get_tensor_dtype("Y:0"), tf.int32)
    self.assertEqual(graph.get_tensor_dtype("Z1:0"), tf.int32)
    self.assertEqual(graph.get_tensor_dtype("Z2:0"), tf.int32)

    self.assertEqual(graph.get_tensor_shape("X:0"), tf.TensorShape([3, 4]))
    self.assertEqual(graph.get_tensor_shape("Y:0"), tf.TensorShape([4, 5]))
    self.assertEqual(graph.get_tensor_shape("Z1:0"), tf.TensorShape([3, 5]))
    self.assertEqual(graph.get_tensor_shape("Z2:0"), tf.TensorShape([3, 5]))

    self.assertEqual(graph.get_tensor_num_entries("X:0"), 12)
    self.assertEqual(graph.get_tensor_num_entries("Y:0"), 20)
    self.assertEqual(graph.get_tensor_num_entries("Z1:0"), 15)
    self.assertEqual(graph.get_tensor_num_entries("Z2:0"), 15)

    graph.set_tensor_final("Z1:0")

    self.assertEqual(graph.compute_memory_contents_under_schedule([0, 1, 2, 3]),
                     [frozenset(["X:0"]), frozenset(["X:0", "Y:0"]),
                      frozenset(["X:0", "Y:0", "Z1:0"]),
                      frozenset(["X:0", "Y:0", "Z1:0", "Z2:0"])])
    self.assertEqual(graph.compute_memory_contents_under_schedule([0, 1, 3, 2]),
                     [frozenset(["X:0"]), frozenset(["X:0", "Y:0"]),
                      frozenset(["X:0", "Y:0", "Z2:0"]),
                      frozenset(["X:0", "Y:0", "Z1:0"])])

  def testTensorFlowGraph(self):
    tf_graph = tf.Graph()
    with tf_graph.as_default():
      with tf.device("/device:CPU:0"):
        x = tf.zeros([3, 4], dtype=tf.int32, name="X")
        y = tf.zeros([4, 5], dtype=tf.int32, name="Y")
        tf.matmul(x, y, name="Z1")
        tf.matmul(x, y, name="Z2")

    graph = graph_interface.GraphInterface(tf_graph,
                                           canonical_device="/device:CPU:0")
    self.VerifyGraphInterface(graph)

    self.assertCountEqual(graph.get_operation_mtf_dimension_names("X"), [])
    self.assertCountEqual(graph.get_operation_mtf_dimension_names("Y"), [])
    self.assertCountEqual(graph.get_operation_mtf_dimension_names("Z1"), [])
    self.assertCountEqual(graph.get_operation_mtf_dimension_names("Z2"), [])

    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("X:0"), [])
    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("Y:0"), [])
    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("Z1:0"), [])
    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("Z2:0"), [])

    self.assertEqual(graph.get_tensor_device("X:0"), "/device:CPU:0")
    self.assertEqual(graph.get_tensor_device("Y:0"), "/device:CPU:0")
    self.assertEqual(graph.get_tensor_device("Z1:0"), "/device:CPU:0")
    self.assertEqual(graph.get_tensor_device("Z2:0"), "/device:CPU:0")

    self.assertTrue(graph.is_tensor_on_canonical_device("X:0"))
    self.assertTrue(graph.is_tensor_on_canonical_device("Y:0"))
    self.assertTrue(graph.is_tensor_on_canonical_device("Z1:0"))
    self.assertTrue(graph.is_tensor_on_canonical_device("Z2:0"))

    self.assertEqual(graph.compute_cost_graph().SerializeToString(),
                     self._cost_graph_string)
    self.assertEqual(graph.compute_cost_graph(devices=["/device:CPU:0"])
                     .SerializeToString(),
                     self._cost_graph_string)
    self.assertEqual(graph.compute_cost_graph(devices=[]).SerializeToString(),
                     self._sizeless_cost_graph_string)

  def testMeshTensorFlowGraph(self):
    mtf_graph = mtf.Graph()
    mesh = mtf.Mesh(mtf_graph, "my_mesh")
    x = mtf.Constant(mesh, 0,
                     shape=mtf.convert_to_shape("a:3,b:4"),
                     dtype=tf.int32,
                     name="X").outputs[0]
    y = mtf.Constant(mesh, 0,
                     shape=mtf.convert_to_shape("b:4,c:5"),
                     dtype=tf.int32,
                     name="Y").outputs[0]
    mtf.EinsumOperation([x, y], mtf.convert_to_shape("a:3,c:5"), name="Z1")
    mtf.EinsumOperation([x, y], mtf.convert_to_shape("a:3,c:5"), name="Z2")

    graph = graph_interface.GraphInterface(mtf_graph)
    self.VerifyGraphInterface(graph)

    self.assertCountEqual(graph.get_operation_mtf_dimension_names("X"),
                          ["a", "b"])
    self.assertCountEqual(graph.get_operation_mtf_dimension_names("Y"),
                          ["b", "c"])
    self.assertCountEqual(graph.get_operation_mtf_dimension_names("Z1"),
                          ["a", "b", "c"])
    self.assertCountEqual(graph.get_operation_mtf_dimension_names("Z2"),
                          ["a", "b", "c"])

    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("X:0"),
                          ["a", "b"])
    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("Y:0"),
                          ["b", "c"])
    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("Z1:0"),
                          ["a", "c"])
    self.assertCountEqual(graph.get_tensor_mtf_dimension_names("Z1:0"),
                          ["a", "c"])

    self.assertIsNone(graph.get_tensor_device("X:0"))
    self.assertIsNone(graph.get_tensor_device("Y:0"))
    self.assertIsNone(graph.get_tensor_device("Z1:0"))
    self.assertIsNone(graph.get_tensor_device("Z2:0"))

    self.assertTrue(graph.is_tensor_on_canonical_device("X:0"))
    self.assertTrue(graph.is_tensor_on_canonical_device("Y:0"))
    self.assertTrue(graph.is_tensor_on_canonical_device("Z1:0"))
    self.assertTrue(graph.is_tensor_on_canonical_device("Z2:0"))

    self.assertEqual(graph.compute_cost_graph().SerializeToString(),
                     self._deviceless_cost_graph_string)
    self.assertEqual(graph.compute_cost_graph(devices=[]).SerializeToString(),
                     self._deviceless_cost_graph_string)

  def testNotAGraph(self):
    self.assertRaises(TypeError, graph_interface.GraphInterface, "hello")


if __name__ == "__main__":
  tf.disable_v2_behavior()
  tf.test.main()



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SIMD Mesh implementation (for TPU/XLA)."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import os

from mesh_tensorflow import ops_with_redefined_builtins as mtf
from mesh_tensorflow import tpu_variables
from mesh_tensorflow import utils
from six.moves import xrange  # pylint: disable=redefined-builtin

import tensorflow.compat.v1 as tf

from tensorflow.python.tpu.ops import tpu_ops  # pylint: disable=g-direct-tensorflow-import


class SimdMeshImpl(mtf.MeshImpl):
  """Mesh implementation for TPU using SIMD and MPI operations."""

  def __init__(self,
               shape,
               layout,
               devices=None,
               device_assignment=None,
               logical_to_physical=None,
               allreduce_in_bfloat16_max_group_size=32,
              ):
    """Create a SimdMeshImpl.

    Args:
      shape: an input to mtf.convert_to_shape()
      layout: an input to mtf.convert_to_layout_rules()
      devices: deprecated
      device_assignment: a tf.tpu.experimental.DeviceAssignment -
        devices must be asssigned in lexicographic order
      logical_to_physical: an optional permutation representing the mapping
        from logical cores to "physical" cores, where the physical cores are
        listed in lexicographic order in the physical mesh, and the logical
        cores are listed in lexicographic order in the logical mesh.
        Default is lexicographic order.
      allreduce_in_bfloat16_max_group_size: an integer.  Allreduces of bfloat16
        tensors are done in float32 if the group size exceeds this value.
    """
    super(SimdMeshImpl, self).__init__(shape, layout)
    if devices is not None:
      tf.logging.warning("SimdMeshImpl ignoring devices %s" % devices)
    self._device_assignment = device_assignment
    tf.logging.info("SimdMeshImpl init: {0} {1}".format(shape, layout))
    tf.logging.info("Device Assignment: {0}".format(device_assignment))
    if logical_to_physical is None:
      # TODO(noam): maybe use auto_logical_to_physical_tpu() here
      logical_to_physical = list(range(self.size))
    if sorted(logical_to_physical) != list(range(self.size)):
      raise ValueError(
          "logical_to_physical must be a permutation on range(shape.size)"
          " shape=%s logical_to_physical=%s" % (shape, logical_to_physical))
    self._logical_to_physical = logical_to_physical
    self._physical_to_logical = [None] * self.size
    for logical, physical in enumerate(self._logical_to_physical):
      self._physical_to_logical[physical] = logical
    self._pnum_tensor = None
    self.graph_device_function_stacks = []
    self.copy_master_to_slice_ops = []
    self._allreduce_in_bfloat16_max_group_size = (
        allreduce_in_bfloat16_max_group_size)

  @property
  def pnum_tensor(self):
    if self._pnum_tensor is not None:
      return self._pnum_tensor
    with utils.outside_all_rewrites():
      tf.logging.info("Create pnum_tensor")
      self._pnum_tensor = tpu_ops.tpu_replicated_input(
          self._physical_to_logical, name="pnum_constants")
      return self._pnum_tensor

  def l2p(self, logical_pnum):
    return self._logical_to_physical[logical_pnum]

  def p2l(self, physical_pnum):
    return self._physical_to_logical[physical_pnum]

  class LaidOutTensor(object):
    """One Slice."""

    def __init__(self, tensor_list):
      assert isinstance(tensor_list, list)
      self._tensor_list = tensor_list

    def __repr__(self):
      return "[" + ",".join([str(t) for t in self._tensor_list]) + "]"

    @property
    def tensor_list(self):
      return self._tensor_list

    @property
    def one_slice(self):
      return self._tensor_list[0]

    @classmethod
    def from_tensor_list(cls, tensor_list):
      return cls(tensor_list)

    @property
    def all_slices(self):
      return self._tensor_list

    @property
    def slice_shape(self):
      return self.one_slice.shape.as_list()

    def to_laid_out_tensor(self):
      return self

  class LaidOutVariable(object):
    """Maintains slice-variables and copy operations."""

    def __init__(self, variable, mesh_impl):
      """Create a LaidOutVariable.

      Args:
        variable: a Variable (Operation)
        mesh_impl: a MeshImpl
      """
      self._variable = variable
      self._mesh_impl = mesh_impl
      shape = variable.outputs[0].shape
      slice_shape = mesh_impl.slice_shape(shape)
      base_name = variable.name
      slices = []
      slices_with_master_dtype = []
      with tf.device(variable.master_device), utils.outside_all_rewrites():
        zero_tensor = tf.zeros(slice_shape, dtype=variable.slice_dtype)

      # pylint: disable=protected-access
      init_device_stack = tf.get_default_graph()._device_function_stack

      if not mesh_impl.graph_device_function_stacks:
        for pnum in xrange(mesh_impl.size):
          tpu_device = mesh_impl.device_assignment.tpu_device(replica=pnum)
          with tf.device(tpu_device):
            mesh_impl.graph_device_function_stacks.append(
                tf.get_default_graph()._device_function_stack.copy())

      for physical_pnum in xrange(mesh_impl.size):
        slice_var_name = base_name + "_slice_%d" % physical_pnum
        # Use tf.Variable instead of tf.get_variable since latter adds lots of
        # useless operations to the TF graph. Use tf.get_variable only if
        # in a AUTO_REUSE scope.
        # Note: Repeatedly 'with tf.device():' slows down the graph
        # construction. Therefore we directly use the cached device_stack here.
        tf.get_default_graph()._device_function_stack = (
            mesh_impl.graph_device_function_stacks[physical_pnum])

        if tf.get_variable_scope().reuse == tf.AUTO_REUSE:
          slice_var = tf.get_variable(
              initializer=zero_tensor,
              trainable=self._variable.trainable,
              collections=["TPU_VAR"],
              dtype=variable.slice_dtype,
              name=slice_var_name)
        else:
          slice_var = tf.Variable(
              initial_value=zero_tensor,
              trainable=self._variable.trainable,
              collections=["TPU_VAR"],
              dtype=variable.slice_dtype,
              name=slice_var_name,
              expected_shape=slice_shape)

        slices.append(slice_var)

      # Restore the initial stack
      tf.get_default_graph()._device_function_stack = init_device_stack
      # pylint: enable=protected-access

      self._laid_out_tensor = mesh_impl.LaidOutTensor(
          [tpu_variables.ReplicatedVariable(base_name, slices)])
      with tf.device(variable.master_device), utils.outside_all_rewrites():
        if os.environ.get("MTF_SEQUENCE_MODE", "") == "1":
          if mesh_impl.copy_master_to_slice_ops:
            with tf.control_dependencies(
                [mesh_impl.copy_master_to_slice_ops[-1]]):
              self._copy_master_to_slices = self._gen_copy_master_to_slices_op(
                  variable.get_master(), shape, slices, slice_shape)
          else:
            self._copy_master_to_slices = self._gen_copy_master_to_slices_op(
                variable.get_master(), shape, slices, slice_shape)

          mesh_impl.copy_master_to_slice_ops.append(self._copy_master_to_slices)
        else:
          self._copy_master_to_slices = self._gen_copy_master_to_slices_op(
              variable.get_master(), shape, slices, slice_shape)
        slices_with_master_dtype = [
            tf.cast(s, variable.master_dtype) for s in slices]
        slices_with_master_dtype = [
            slices_with_master_dtype[mesh_impl.l2p(logical_pnum)]
            for logical_pnum in range(mesh_impl.size)]
        self._copy_slices_to_master = variable.assign_to_master(
            mesh_impl.combine_slices(slices_with_master_dtype, shape,
                                     device=variable.master_device))

    def _gen_copy_master_to_slices_op(self, master_variable, master_shape,
                                      slices, slice_shape):
      """Generate ops which slices master and assign to slices.

      Args:
        master_variable: The master variable.
        master_shape: The shape of master variable.
        slices: The list of slice-variables in physical order.
        slice_shape: The shape of the slice variable.
      Returns:
        A grouped tf.assign ops.
      """
      mesh_impl = self._mesh_impl
      master_layout = mesh_impl.tensor_layout(master_shape)
      # For handling case: master is float32 and slices are bfloat16.
      if master_variable.dtype != slices[0].dtype:
        master_variable = tf.cast(master_variable, slices[0].dtype)
      assign_ops = []
      if master_layout.is_fully_replicated:
        assign_ops = [tf.assign(t, master_variable) for t in slices]
      else:
        slice_dict = {}
        for logical_pnum in xrange(len(slices)):
          slice_begin = mesh_impl.slice_begin(master_shape, logical_pnum)
          slice_begin_tuple = tuple(slice_begin)
          # Reuse the same slice if slice_begin doesn't change.
          if slice_begin_tuple not in slice_dict:
            slice_dict[slice_begin_tuple] = tf.slice(master_variable,
                                                     slice_begin, slice_shape)
          physical_pnum = mesh_impl.l2p(logical_pnum)
          assign_ops.append(
              tf.assign(slices[physical_pnum], slice_dict[slice_begin_tuple]))
      return tf.group(assign_ops)

    def assign_to_slices(self, assign_fn, values, assign_to_tensor_list=None):
      """Assign to the slice variables.

      Args:
        assign_fn: a function from
          (mtf.Variable, tf.Variable, tf.Tensor) -> tf.Operation
        values: a list of tf.Tensor
        assign_to_tensor_list: an optional list of tf.Variable

      Returns:
        a tf.operation
      """
      if assign_to_tensor_list is None:
        assign_to_tensor_list = self._laid_out_tensor.all_slices
      # Handle both N -> 1 and N -> N cases.
      num_slices = min(len(assign_to_tensor_list), len(values))
      devices = [""] * num_slices
      return tf.group(
          mtf.parallel(devices, assign_fn,
                       [self._variable] * len(devices),
                       assign_to_tensor_list[:num_slices],
                       values[:num_slices]))

    @property
    def laid_out_tensor(self):
      return self._laid_out_tensor

    @property
    def copy_master_to_slices(self):
      return self._copy_master_to_slices

    @property
    def copy_slices_to_master(self):
      return self._copy_slices_to_master

  def laid_out_pnum(self):
    """Returns a LaidOutTensor containing the logical processor number.

    Returns:
      a LaidOutTensor where each slice is an integer scalar
    """
    return self.LaidOutTensor([self.pnum_tensor])

  def _create_group_assignment(self, mesh_axes):
    """Create group assignment for XLA cross replica ops (physical pnums)."""

    partitioning = {}
    for logical_pnum in xrange(self.size):
      group = mtf.pnum_to_group(self.shape, mesh_axes, logical_pnum)
      if group not in partitioning:
        partitioning[group] = []
      partitioning[group].append(self.l2p(logical_pnum))
    group_assignment = []
    for group, physical_pnums in partitioning.items():
      group_assignment.append(physical_pnums)
    return group_assignment

  def allreduce(self, x, mesh_axes, reduction_fn_string):
    """Grouped allreduce, (summed across the given dimensions).

    Args:
      x: a LaidOutTensor
      mesh_axes: a list of integers
      reduction_fn_string: "SUM"
    Returns:
      a LaidOutTensor
    Raises:
      ValueError: if the reduction is not yet implemented.
    """
    if not mesh_axes:
      return x
    x = x.to_laid_out_tensor()
    if reduction_fn_string == "SUM":
      group_assignment = self._create_group_assignment(mesh_axes)
      group_size = len(group_assignment[0])
      tf_in = x.one_slice
      dtype = tf_in.dtype
      if dtype == tf.float32:
        cast_to_float32 = False
      elif dtype == tf.bfloat16:
        cast_to_float32 = (
            group_size > self._allreduce_in_bfloat16_max_group_size)
      else:
        tf.logging.info("Casting %s to float32 for allreduce" % tf_in.dtype)
        cast_to_float32 = True
      if cast_to_float32:
        tf_in = tf.cast(tf_in, tf.float32)
      tf_out = tpu_ops.cross_replica_sum(tf_in, group_assignment)
      if cast_to_float32:
        tf_out = tf.cast(tf_out, dtype)
      return self.LaidOutTensor([tf_out])
    else:
      for axis in mesh_axes:
        x = self.allconcat(x, axis, 0, stack=True)
        x = self.LaidOutTensor(
            [mtf.reduction_fn(reduction_fn_string)(x.one_slice, 0)])
      return x

  def allconcat(self, x, mesh_axis, concat_axis, stack=False):
    """Grouped allconcat (like MPI allgather followed by concat).

    TODO(noam): inefficient - replace with a XLA allconcat when available

    Args:
      x: a LaidOutTensor
      mesh_axis: an integer - the mesh axis along which to group
      concat_axis: an integer (the Tensor axis along which to concatenate)
      stack: a boolean - whether to stack instead of concat
    Returns:
      a LaidOutTensor
    """
    x = x.to_laid_out_tensor()
    coord = self.laid_out_pcoord(mesh_axis)
    t = x.one_slice
    old_shape = t.shape.as_list()
    num_parts = self.shape[mesh_axis].size
    t = tf.expand_dims(t, concat_axis)
    t *= tf.reshape(
        tf.one_hot(coord.one_slice, num_parts, dtype=t.dtype),
        [num_parts if i == concat_axis else 1
         for i in xrange(len(old_shape) + 1)])
    if not stack:
      new_shape = old_shape[:]
      new_shape[concat_axis] *= num_parts
      t = tf.reshape(t, new_shape)
    return self.allreduce(self.LaidOutTensor([t]), [mesh_axis], "SUM")

  def alltoall(self, x, mesh_axis, split_axis, concat_axis):
    """Grouped alltoall (like MPI alltoall with splitting and concatenation).

    Args:
      x: a LaidOutTensor
      mesh_axis: an integer the mesh axis along which to group
      split_axis: an integer (the Tensor axis along which to split)
      concat_axis: an integer (the Tensor axis along which to concatenate)
    Returns:
      a LaidOutTensor
    """
    x = x.to_laid_out_tensor()
    t = x.one_slice
    group_assignment = self._create_group_assignment([mesh_axis])
    dtype = t.dtype
    if dtype == tf.float32:
      # There seems to be a bug with float32 alltoall.
      # Do it in bfloat16 until the bug is fixed.
      # TODO(noam): file a bug
      t = tf.to_bfloat16(t)
    t = tpu_ops.all_to_all(
        t,
        concat_dimension=concat_axis,
        split_dimension=split_axis,
        split_count=len(group_assignment[0]),
        group_assignment=group_assignment)
    t = tf.cast(t, dtype)
    x = self.LaidOutTensor([t])
    return x

  def receive(self, x, mesh_axis, source_pcoord):
    """Collective receive in groups.

    Each group contains the processors that differ only in mesh_axis.

    ```python
    group_size = self.shape[mesh_axis].size
    ```

    Args:
      x: a LaidOutTensor
      mesh_axis: an integer
      source_pcoord: a list of optional integers. Each element is either None
        or an integer in [0, group_size). If source_pcoord[k] is None, then the
        output for the k-th processor in each group is a zero tensor. If
        source_pcoord[k] is not None, then the output for the k-th processor in
        each group is equal to the input for the source_pcoord[k]-th processor
        in that group.

    Returns:
      a LaidOutTensor
    """
    x = x.to_laid_out_tensor()
    t = x.one_slice
    source_target_pairs = []

    for pnum in xrange(self.size):
      coord = mtf.pnum_to_processor_coordinates(self.shape, pnum)
      k = coord[mesh_axis]
      if source_pcoord[k] is not None:
        coord[mesh_axis] = source_pcoord[k]
        source_pnum = mtf.processor_coordinates_to_pnum(self.shape, coord)
        source_target_pairs.append(
            [self.l2p(source_pnum),
             self.l2p(pnum)])

    if not source_target_pairs:
      ret = tf.zeros_like(t, t.dtype)
    elif t.dtype in [tf.float32, tf.bfloat16, tf.int32]:
      ret = tpu_ops.collective_permute(t, source_target_pairs)
    else:
      # If t is not one of the allowed types, cast and cast back.
      ret = tf.cast(tpu_ops.collective_permute(
          tf.cast(t, tf.float32), source_target_pairs), t.dtype)

    return self.LaidOutTensor([ret])

  def slice(self, tf_tensor, tensor_shape):
    """"Slice out the corresponding part of tensor given the pnum variable."""
    tensor_layout = self.tensor_layout(tensor_shape)

    if tensor_layout.is_fully_replicated:
      return self.LaidOutTensor([tf_tensor])
    else:
      slice_shape = self.slice_shape(tensor_shape)
      slice_begins = [
          self.slice_begin(tensor_shape, pnum) for pnum in xrange(self.size)
      ]
      slice_begins_tensor = tf.stack(slice_begins)
      # slice on source device
      selected_slice_begin = tf.gather(slice_begins_tensor, self.pnum_tensor)
      return self.LaidOutTensor(
          [tf.slice(tf_tensor, selected_slice_begin, slice_shape)])

  def slicewise(self, fn, *inputs):
    """Execute a function in parallel on all slices.

    Args:
      fn: a function from tf.Tensors to tf.Tensor or a tuple of tf.Tensors.
      *inputs: a list of inputs.  Each input is either a LaidOutTensor or
        is convertible to a tf.Tensor.
    Returns:
      a LaidOutTensor, or a tuple of LaidOutTensors if fn returns a tuple.
    """
    if fn == tf.add:
      assert len(inputs) == 2
      if isinstance(inputs[0], mtf.LazyAllreduceSum):
        # sum of LazyAllreduceSum (keep delaying the allreduce)
        return inputs[0] + inputs[1]
    # convert all inputs to LaidOutTensor where possible
    inputs = mtf.convert_args_to_laid_out_tensors(inputs)
    ret = fn(*[
        x.one_slice if isinstance(x, self.LaidOutTensor) else x
        for x in inputs])
    if isinstance(ret, tuple):
      return tuple([self.LaidOutTensor([t]) for t in ret])
    else:
      return self.LaidOutTensor([ret])

  @property
  def device_assignment(self):
    return self._device_assignment

  @property
  def devices(self):
    return self._devices

  def random(self, shape, tf_fn, kwargs):
    """Call a random tf operation (e.g. random_uniform).

    Args:
      shape: a Shape
      tf_fn: a function such as tf.random.uniform
      kwargs: kwargs to pass to tf_fn, except for seed

    Returns:
      a LaidOutTensor
    """
    # TODO(noam): can we make things better with stateless_random?
    slice_shape = self.slice_shape(shape)
    x = tf_fn(slice_shape, **kwargs)
    # TPU does not have seeds enabled.  Sync up the
    # random choices by zeroing out all but the first core per group of
    # identical slices, then allreducing by group.
    layout = self.tensor_layout(shape)
    # we need to sync across these axes.
    mesh_axes = [i for i in xrange(self.ndims)
                 if i not in layout.tensor_axis_to_mesh_axis]
    multiplier = 1.0
    for axis in mesh_axes:
      multiplier *= tf.cast(
          tf.equal(self.laid_out_pcoord(axis).one_slice, 0), x.dtype)
    x *= multiplier
    x = self.LaidOutTensor([x])
    x = self.allreduce(x, mesh_axes, "SUM")
    return x

  def export_to_tf_tensor(self, x, laid_out_x):
    """Turn a Tensor into a tf.Tensor.

    Args:
      x: a Tensor
      laid_out_x: a LaidOutTensor
    Returns:
      a tf.Tensor
    """
    tensor_layout = self.tensor_layout(x.shape)
    if not tensor_layout.is_fully_replicated:
      raise NotImplementedError(
          "SimdMeshImpl only supports export_to_tf_tensor of fully-replicated "
          "Tensors.  Try reshaping to new dimension names. "
          " x.shape = %s tensor_layout=%s"
          % (x.shape, tensor_layout))
    return laid_out_x.one_slice

  def import_tf_tensor(self, x, tf_x):
    """Import a tf.Tensor, producing a LaidOutTensor.

    Args:
      x: a Tensor
      tf_x: a tf.Tensor
    Returns:
      a LaidOutTensor
    """
    return self.slice(tf_x, x.shape)

  @property
  def supports_control_dependencies(self):
    return False

  def einsum(self, equation, *slices):
    """Override this for custom einsum implementation.

    Args:
      equation: a string
      *slices: a list of tf.Tensor
    Returns:
      a tf.Tensor
    """
    return tf.einsum(equation, *slices)


def _ring_2d(m, n):
  """Ring-order of a mxn mesh.

  If m and n are both even, then we generate a ring like this:

     0 -- 1 -- 2 -- 3
     |    |    |    |
     15-- 6 -- 5 -- 4
     |    |    |    |
     14-- 7 -- 8 -- 9
     |    |    |    |
     13-- 12-- 11-- 10

  Args:
    m: an integer
    n: an integer
  Returns:
    a list of mxn pairs
  """
  if m == 1:
    return [(0, i) for i in range(n)]
  if n == 1:
    return [(i, 0) for i in range(m)]
  if m % 2 != 0:
    tf.logging.warning("Odd dimension")
    return [(i % m, i // m) for i in range(n * m)]
  ret = [(0, 0)]
  for i in range(m // 2):
    for j in range(1, n):
      ret.append((2 * i, j))
    for j in range(n-1, 0, -1):
      ret.append((2 * i + 1, j))
  for i in range(m-1, 0, -1):
    ret.append((i, 0))
  return ret


def tile_2d(physical_shape, tile_shape,
            outer_name="outer",
            inner_name="inner",
            cores_name=None):
  """2D tiling of a 3d physical mesh.

  The "outer" mesh dimension corresponds to which tile.
  The "inner" mesh dimension corresponds to the position within a tile
  of processors.

  Optionally, if cores_name is specified, then a 3 dimensional logical mesh
  is returned, with the third dimension representing the two different
  cores within a chip.  If cores_name is not specified, then the
  cores-in-a-chip dimension is folded into the inner dimension.

  TODO(noam): explain this better.

  Example:

  tile_2d(physical_shape=[8, 16, 2], tile_shape=[4, 4])

  The "inner" dimension has size 4x4x2=32 and corresponds to the position
  within a 4x4 tile of processors.

  The "outer" dimension has size 8/4 * 16/4 = 8, and corresponds to the 8
  tiles in the mesh.

  Args:
    physical_shape: a triple of integers [X, Y, cores]
    tile_shape: a pair
    outer_name: a string
    inner_name: a string
    cores_name: an optional string

  Returns:
    mesh_shape: a mtf.Shape
    logical_to_physical: a list
  """
  logical_to_physical = []
  p0, p1, p2 = physical_shape
  t0, t1 = tile_shape
  tile_ring = _ring_2d(t0, t1)
  tiles_ring = _ring_2d(p0 // t0, p1 // t1)
  for logical_pnum in range(p0 * p1 * p2):
    logical_tile_num = logical_pnum // (t0 * t1 * p2)
    if p2 == 2 and not cores_name:
      # Go through all chips using core 0, then go through all chips
      #   backwards using core 1.  This is better in the case where
      #   one of the tile dimensions is 1, so the last chip is not adjacent
      #   to the first chip.
      core_in_tile = logical_pnum % (t0 * t1 * p2)
      core_on_chip = core_in_tile // (t0 * t1)
      if core_on_chip == 0:
        logical_pos_in_tile = core_in_tile
      else:
        logical_pos_in_tile = t0 * t1 * p2 - 1 - core_in_tile
    else:
      # Go through all chips once, using both cores on each chip.
      core_on_chip = logical_pnum % p2
      logical_chip_num = logical_pnum // p2
      logical_pos_in_tile = logical_chip_num % (t0 * t1)
    tile_i, tile_j = tile_ring[logical_pos_in_tile]
    tiles_i, tiles_j = tiles_ring[logical_tile_num]
    physical_pnum = core_on_chip + p2 * (
        tile_i * p1 + tile_j +
        tiles_i * p1 * t0 + tiles_j * t1)
    logical_to_physical.append(physical_pnum)
  assert sorted(logical_to_physical) == list(range(p0 * p1  * p2))
  tile_size = t0 * t1 * p2
  num_tiles = p0 * p1 // (t0 * t1)
  if cores_name:
    mesh_shape = mtf.Shape(
        [mtf.Dimension(outer_name, int(num_tiles)),
         mtf.Dimension(inner_name, int(t0 * t1)),
         mtf.Dimension(cores_name, int(p2))])
  else:
    mesh_shape = mtf.Shape(
        [mtf.Dimension(outer_name, int(num_tiles)),
         mtf.Dimension(inner_name, int(tile_size))])
  return mesh_shape, logical_to_physical


def auto_logical_to_physical_tpu(logical_shape,
                                 physical_shape,
                                 return_coordinates=False):
  """Set up a mapping from logical to physical cores for TPU.

  We will try to set up a mapping so that allreduce operations are relatively
  fast, prioritizing the later dimensions in the mesh_shape.

  Example:

  auto_logical_to_physical_tpu(logical_shape=[16, 8], physical_shape=[8, 8, 2])

  Heuristics in this function subject to change.

  Args:
    logical_shape: a list of integers
    physical_shape: a list of integers - typically [X, Y, cores]
    return_coordinates: a boolean - return a list of integer lists (coordinates)
       instead of a list of processor indices

  Returns:
    logical_to_physical: a permutation of range(product(physical_shape)))
  """
  if mtf.list_product(logical_shape) != mtf.list_product(physical_shape):
    raise ValueError(
        "physical and logical shapes must have the same product "
        "physical_shape=%s logical_shape=%s" % (physical_shape, logical_shape))
  # drop logical dimensions of size 1
  logical_shape = [i for i in logical_shape if i != 1]
  num_cores = mtf.list_product(logical_shape)
  # For physical shapes different from what we are used to [2^a, 2^b, 2],
  #   return a simple default value (a lexicographic ordering)
  def _default_value():
    default = list(range(num_cores))
    if return_coordinates:
      default = [mtf.pnum_to_processor_coordinates(i) for i in default]
    return default
  if len(physical_shape) != 3:
    return _default_value()
  p0, p1, p2 = physical_shape
  if p2 != 2:
    return _default_value
  for dimsize in [p0, p1]:
    # if dimsize not a power of 2, give up
    if dimsize & (dimsize - 1):
      return _default_value()
  # At this point, the physical shape has at least 1x1x2=2 cores, so there
  #   must be at least one logical dimension.
  assert logical_shape
  if len(logical_shape) == 1:
    # ring of p0 x p1 chips
    ring = _ring_2d(p0, p1)
    logical_to_physical = []
    for logical_pnum in range(num_cores):
      # Go through all chips using core 0, then go through all chips
      #   backwards using core 1.  This is better in the case where
      #   one of the tile dimensions is 1, so the last chip is not adjacent
      #   to the first chip.
      core_on_chip = logical_pnum // (p0 * p1)
      if core_on_chip == 0:
        chip_num = logical_pnum
      else:
        chip_num = num_cores - 1 - logical_pnum
      i, j = ring[chip_num]
      logical_to_physical.append((i, j, core_on_chip))
  else:
    # We have a p0 x p1 rectangle of chips, which we will tile with rectangular
    #   tiles.  The first logical dimension correspond to the number of tiles,
    #   and the other logical dimensions will correspond to position within a
    #   tile.
    num_tiles = logical_shape[0]
    tile_chips = num_cores // num_tiles // p2
    # If we can, we make each tile occupy exactly one row or column of chips.
    # Otherwise, we make each tile approximately square.
    if len(logical_shape) == 2 and tile_chips == p0:
      t0, t1 = [tile_chips, 1]
    elif len(logical_shape) == 2 and tile_chips == p1:
      t0, t1 = [1, tile_chips]
    else:
      # try to make the tile approximately square
      lg_tile_chips = int(math.log(tile_chips, 2))
      t0 = 2 ** (lg_tile_chips // 2)
      # make sure that the tile fits in the mesh - i.e.
      #   t0 <= p0
      #   t1 == tile_chips // t0 <= p1
      t0 = min(t0, p0)
      t0 = max(t0, tile_chips // p1)
      t1 = tile_chips // t0
    # recursive call to find mapping for one tile
    tile_logical_to_physical = auto_logical_to_physical_tpu(
        logical_shape[1:], [t0, t1, p2], return_coordinates=True)
    tiles_ring = _ring_2d(p0 // t0, p1 // t1)
    logical_to_physical = []
    for logical_pnum in range(num_cores):
      logical_tile_num = logical_pnum // (t0 * t1 * p2)
      logical_pos_in_tile = logical_pnum % (t0 * t1 * p2)
      logical_to_physical.append((
          tiles_ring[logical_tile_num][0] * t0 +
          tile_logical_to_physical[logical_pos_in_tile][0],
          tiles_ring[logical_tile_num][1] * t1 +
          tile_logical_to_physical[logical_pos_in_tile][1],
          tile_logical_to_physical[logical_pos_in_tile][2]))
  if return_coordinates:
    return logical_to_physical
  else:
    return [mtf.processor_coordinates_to_pnum(physical_shape, coord)
            for coord in logical_to_physical]



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Wrapper around vocabulary from the Tensor2Tensor library.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import gin

from mesh_tensorflow.transformer import vocabulary


class T2tVocabulary(vocabulary.Vocabulary):
  """Wrapper around tensor2tensor SubwordTextEncoder.

  1 is already reserved for EOS - no need to shift.
  """

  def __init__(self, filepath):
    """Create a T2tVocabulary.

    Args:
      filepath: a string
    """
    # Only import tensor2tensor if necessary.
    from tensor2tensor.data_generators import text_encoder   # pylint: disable=g-import-not-at-top
    from tensor2tensor.data_generators.ops import subword_text_encoder_ops   # pylint: disable=g-import-not-at-top

    self._filepath = filepath
    self._subword_text_encoder = text_encoder.SubwordTextEncoder(filepath)
    self._subword_text_encoder_encode = (
        subword_text_encoder_ops.subword_text_encoder_encode)

  @property
  def vocab_size(self):
    """Number of ids (including 0=PAD and 1=EOS).

    Returns:
      an integer
    """
    return self._subword_text_encoder.vocab_size

  def encode(self, s):
    """Encode a python string as a list of integers.

    Args:
      s: a string
    Returns:
      a list of integers (not terminated by EOS)
    """
    return self._subword_text_encoder.encode(s)

  def decode(self, ids):
    """Decode a list of integers to a python string.

    Args:
      ids: a list of integers (not terminated by EOS)
    Returns:
      a string
    """
    return self._subword_text_encoder.decode(ids)

  def encode_tf(self, s):
    """Encode a tf.Scalar string to a tf.Tensor.

    This will be necessary for on-the-fly tokenization.

    Args:
      s: a tf.Scalar with dtype tf.string
    Returns:
      a 1d tf.Tensor with dtype tf.int32
    """
    ids = self._subword_text_encoder_encode(s, self._filepath)
    # the c++ op apppends 1=EOS - drop it.
    return ids[:-1]


@gin.configurable
def get_t2t_vocabulary(data_dir=gin.REQUIRED,
                       vocabulary_filename=gin.REQUIRED):
  return T2tVocabulary(os.path.join(data_dir, vocabulary_filename))



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized

import mesh_tensorflow as mtf
from mesh_tensorflow.transformer import utils

import tensorflow.compat.v1 as tf

tf.disable_v2_behavior()


class UtilsTest(parameterized.TestCase, tf.test.TestCase):

  def testDynamicText2self(self):
    batch = 2
    length = 5
    input_tensors = {
        "inputs": [
            [3, 1, 4, 1, 0],
            [1, 4, 3, 2, 1]],
        "inputs_segmentation": [
            [1, 1, 2, 2, 0],
            [1, 2, 2, 2, 2]],
        "inputs_position": [
            [0, 1, 0, 1, 0],
            [0, 0, 1, 2, 3]],
        "targets": [
            [1, 1, 0, 0, 0],
            [9, 8, 1, 2, 1]],
        "targets_segmentation": [
            [1, 2, 0, 0, 0],
            [1, 1, 1, 2, 2]],
        "targets_position": [
            [0, 0, 0, 0, 0],
            [0, 1, 2, 0, 1]]
    }
    expected_output_tensors = {
        "targets": [
            [3, 1, 1, 4, 1, 1, 0, 0, 0, 0],
            [1, 9, 8, 1, 4, 3, 2, 1, 2, 1]],
        "targets_segmentation": [
            [1, 1, 1, 2, 2, 2, 0, 0, 0, 0],
            [1, 1, 1, 1, 2, 2, 2, 2, 2, 2]],
        "targets_position": [
            [0, 1, 2, 0, 1, 2, 0, 0, 0, 0],
            [0, 1, 2, 3, 0, 1, 2, 3, 4, 5]]
    }
    graph = mtf.Graph()
    mesh = mtf.Mesh(graph, "my_mesh")
    batch_dim = mtf.Dimension("batch", batch)
    length_dim = mtf.Dimension("length", length)

    input_shape = mtf.Shape([batch_dim, length_dim])
    mtf_features = {k: mtf.import_tf_tensor(mesh, v, input_shape)
                    for k, v in input_tensors.items()}
    mtf_outputs = utils._dynamic_text2self(mtf_features)
    mesh_impl = mtf.placement_mesh_impl.PlacementMeshImpl(
        shape=[], layout={}, devices=[""])
    lowering = mtf.Lowering(graph, {mesh: mesh_impl})
    for k, v in expected_output_tensors.items():
      out = lowering.export_to_tf_tensor(mtf_outputs[k])
      actual = self.evaluate(out)
      self.assertAllEqual(actual, v)

  def testCleanDecodes(self):
    cleaned_decodes = utils.clean_decodes([
        [2, 0, 2, 1, 3, 2, 0],
        [1, 2, 2, 2, 2, 2, 2],
        [2, 2, 1, 1, 1, 1, 1],
        [2, 2, 2, 2, 2, 2, 2]
    ])
    with self.test_session() as sess:
      self.assertAllEqual(
          sess.run(cleaned_decodes),
          [
              [2, 0, 2, 1, 0, 0, 0],
              [1, 0, 0, 0, 0, 0, 0],
              [2, 2, 1, 0, 0, 0, 0],
              [2, 2, 2, 2, 2, 2, 2]
          ])

if __name__ == "__main__":
  tf.test.main()



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of beam search with penalties."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gin
from mesh_tensorflow import ops_with_redefined_builtins as mtf
import tensorflow.compat.v1 as tf

# Assuming EOS_ID is 1
EOS_ID = 1
# Default value for INF
INF = 1. * 1e7


def compute_topk_scores_and_seq(sequences, scores, scores_to_gather, flags,
                                beam_dim, prefix="default"):
  """Given sequences and scores, will gather the top k=beam size sequences.

  This function is used to grow alive, and finished. It takes sequences,
  scores, and flags, and returns the top k from sequences, scores_to_gather,
  and flags based on the values in scores.

  This method permits easy introspection using tfdbg.  It adds two named ops
  that are prefixed by `prefix`:
    - _topk_seq: the tensor for topk_seq returned by this method.
    - _topk_flags: the tensor for topk_finished_flags returned by this method.

  Args:
    sequences: Tensor of sequences that we need to gather from.
      [batch_size, beam_size, seq_length]
    scores: Tensor of scores for each sequence in sequences.
      [batch_size, beam_size]. We will use these to compute the topk.
    scores_to_gather: Tensor of scores for each sequence in sequences.
      [batch_size, beam_size]. We will return the gathered scores from here.
      Scores to gather is different from scores because for grow_alive, we will
      need to return log_probs, while for grow_finished, we will need to return
      the length penalized scores.
    flags: Tensor of bools for sequences that say whether a sequence has reached
      EOS or not
    beam_dim: mtf.Dimension
    prefix: an optional string
  Returns:
    Tuple of
    (topk_seq [batch_size, beam_size, decode_length],
     topk_gathered_scores [batch_size, beam_size],
     topk_finished_flags[batch_size, beam_size],
     selector)
  """
  unused_batch_dim, old_beam_dim, unused_length_dim = sequences.shape.dims
  _, topk_indices = mtf.top_k(scores, old_beam_dim, k_dim=beam_dim)

  selector = mtf.one_hot(topk_indices, old_beam_dim, dtype=tf.float32)

  # Gather up the highest scoring sequences.
  # For each operation added, give it
  # a concrete name to simplify observing these operations with tfdbg.
  # Clients can capture these tensors by watching these node names.
  def gather(tensor, name):
    with tf.name_scope(prefix + name):
      output_shape = mtf.Shape(
          [beam_dim if d == old_beam_dim else d for d in tensor.shape.dims])
      return mtf.gather(
          tensor, topk_indices, old_beam_dim, output_shape=output_shape)
  topk_seq = gather(sequences, "_seq")
  topk_flags = gather(flags, "_flags")
  topk_gathered_scores = gather(scores_to_gather, "_scores")
  return topk_seq, topk_gathered_scores, topk_flags, selector


@gin.configurable
def beam_search(logits_fn,
                initial_ids,
                alpha,
                states=None,
                eos_id=EOS_ID,
                stop_early=True,
                decode_length=None,
                use_tpu=True,
                dtype=tf.float32,
                layout=None,
                mesh_shape=None,
                num_prefilter=2):
  """Beam search with length penalties.

  Requires a function that can take the currently decoded symbols and return
  the logits for the next symbol. The implementation is inspired by
  https://arxiv.org/abs/1609.08144.

  When running, the beam search steps can be visualized by using tfdbg to watch
  the operations generating the output ids for each beam step.  These operations
  have the pattern:
    (alive|finished)_topk_(seq,scores)

  Operations marked `alive` represent the new beam sequences that will be
  processed in the next step.  Operations marked `finished` represent the
  completed beam sequences, which may be padded with 0s if no beams finished.

  Operations marked `seq` store the full beam sequence for the time step.
  Operations marked `scores` store the sequence's final log scores.

  The beam search steps will be processed sequentially in order, so when
  capturing observed from these operations, tensors, clients can make
  assumptions about which step is being recorded.

  num_prefilter is a theoretically lossy shortcut around slow performance of
  top_k on TPU on large Tensors and large k.  This option should be removed once
  better top_k implementations on TPU are avialable.  If num_prefilter is set to
  a nonzero value, then at each step we first compute the top num_prefilter
  sequences per beam and then compute the top k sequences overall from among
  those.  Empirically, there seems to be no quality difference in setting
  num_prefilter to 2.

  Args:
    logits_fn: Interface to the model, to provide logits.
        Should take:
          step_num - mtf Scalar
          ids - mtf Tensor with shape [batch, beam, length]
        Should return:
          logits - [batch, beam, vocab_size], dtype=dtype
    initial_ids: a mtf.Tensor with shape [batch_dim, beam_dim, length_dim])
    alpha: alpha for length penalty.
    states: list of mtf.Tensor
    eos_id: ID for end of sentence.
    stop_early: a boolean - stop once best sequence is provably determined.
    decode_length: a mtf Scalar of dtype tf.int32 - maximum length of decodes
    use_tpu: a boolean
    dtype: a tf.dtype
    layout: an optional string
    mesh_shape: an optional string
    num_prefilter: an optional integer
  Returns:
    Tuple of
    (decoded beams [batch, beam, length]
     decoding probabilities [batch, beam_size])
  """
  batch_dim, beam_dim, length_dim = initial_ids.shape.dims
  batch_and_beam_dim = mtf.Dimension(
      batch_dim.name, batch_dim.size * beam_dim.size)
  mesh = initial_ids.mesh

  batch_by_beam = mtf.Shape([batch_dim, beam_dim])
  initial_log_probs = mtf.broadcast(
      mtf.one_hot(
          mtf.constant(mesh, 0, dtype=tf.int32),
          beam_dim,
          on_value=0.0,
          off_value=-INF,
          dtype=dtype),
      batch_by_beam)

  length_scalar = mtf.constant(mesh, length_dim.size, dtype=tf.int32)
  if decode_length is None:
    decode_length = length_scalar
  else:
    decode_length = mtf.minimum(decode_length, length_scalar)

  alive_log_probs = initial_log_probs
  alive_seq = initial_ids

  # Finished will keep track of all the sequences that have finished so far
  # Finished log probs will be negative infinity in the beginning
  # finished_flags will keep track of booleans
  finished_seq = initial_ids
  finished_scores = mtf.constant(mesh, -INF, batch_by_beam, dtype=dtype)

  # Setting the scores of the initial to negative infinity.
  finished_flags = mtf.constant(mesh, False, batch_by_beam, tf.bool)

  def grow_finished(finished_seq, finished_scores, finished_flags, curr_seq,
                    curr_scores, curr_finished):
    """Given sequences and scores, will gather the top k=beam size sequences.

    Args:
      finished_seq: Current finished sequences.
        [batch, beam, length]
      finished_scores: scores for each of these sequences.
        [batch, beam]
      finished_flags: finished bools for each of these sequences.
        [batch, beam]
      curr_seq: current topk sequence that has been grown by one position.
        [batch, beam, length]
      curr_scores: scores for each of these sequences. [batch, beam]
      curr_finished: Finished flags for each of these sequences.
        [batch, beam]
    Returns:
      Tuple of
        (Topk sequences based on scores,
         log probs of these sequences,
         Finished flags of these sequences,
         None (no states))
    """

    # Set the scores of the unfinished seq in curr_seq to large negative
    # values
    curr_scores += (1. - mtf.cast(curr_finished, curr_scores.dtype)) * -INF
    unused_batch_dim, beam_dim, unused_length_dim = finished_seq.shape.dims
    # concatenating the sequences and scores along beam axis
    def _my_concat(a, b):
      a = mtf.rename_dimension(a, "beam", "triple_beam")
      b = mtf.rename_dimension(b, "double_beam", "triple_beam")
      return mtf.concat([a, b], "triple_beam")

    curr_finished_seq = _my_concat(finished_seq, curr_seq)
    curr_finished_scores = _my_concat(finished_scores, curr_scores)
    curr_finished_flags = _my_concat(finished_flags, curr_finished)
    return compute_topk_scores_and_seq(
        curr_finished_seq, curr_finished_scores, curr_finished_scores,
        curr_finished_flags, beam_dim, "grow_finished")

  def grow_alive(curr_seq, curr_scores, curr_log_probs, curr_finished):
    """Given sequences and scores, will gather the top k=beam size sequences.

    Args:
      curr_seq: current topk sequence that has been grown by one position.
        [batch, beam, length]
      curr_scores: scores for each of these sequences. [batch_size, beam_size]
      curr_log_probs: log probs for each of these sequences.
        [batch, beam]
      curr_finished: Finished flags for each of these sequences.
        [batch, beam]
    Returns:
      Tuple of
        (Topk sequences based on scores,
         log probs of these sequences,
         Finished flags of these sequences)
    """
    # Set the scores of the finished seq in curr_seq to large negative
    # values
    curr_scores += mtf.cast(curr_finished, curr_scores.dtype) * -INF
    return compute_topk_scores_and_seq(curr_seq, curr_scores, curr_log_probs,
                                       curr_finished, beam_dim,
                                       "grow_alive")

  def grow_topk(i, alive_seq, alive_log_probs, states=None):
    r"""Inner beam search loop.

    This function takes the current alive sequences, and grows them to topk
    sequences where k = 2*beam. We use 2*beam because, we could have beam_size
    number of sequences that might hit <EOS> and there will be no alive
    sequences to continue. With 2*beam_size, this will not happen. This relies
    on the assumption the vocab size is > beam size. If this is true, we'll
    have at least beam_size non <EOS> extensions if we extract the next top
    2*beam words.
    Length penalty is given by = (5+len(decode)/6) ^ -\alpha. Pls refer to
    https://arxiv.org/abs/1609.08144.

    Args:
      i: loop index
      alive_seq: Topk sequences decoded so far [batch, beam, length]
      alive_log_probs: probabilities of these sequences. [batch, beam]
      states: optional list of mtf.Tensor
    Returns:
      Tuple of
        (Topk sequences extended by the next word,
         The log probs of these sequences,
         The scores with length penalty of these sequences,
         Flags indicating which of these sequences have finished decoding,
         list of transformed decoding states)
    """
    logits, new_states = logits_fn(i, alive_seq, states)
    batch_dim, beam_dim, vocab_dim = logits.shape.dims

    # Convert logits to normalized log probs
    candidate_log_probs = mtf.log_softmax(logits, vocab_dim)

    # Multiply the probabilities by the current probabilities of the beam.
    # (batch_size, beam_size, vocab_size) + (batch_size, beam_size, 1)
    log_probs = candidate_log_probs + alive_log_probs

    length_penalty = mtf.pow(((5. + mtf.cast(i + 1, logits.dtype)) / 6.), alpha)

    # scores have shape [batch, beam, vocab]
    curr_scores = log_probs / length_penalty

    # We find the top 2k sequences to make sure we get k alive sequences.
    #
    # TODO(noam): This is inefficient.  We should separately compute the k
    # finished sequences (previously alive sequences + EOS), and the top k new
    # alive sequences.
    double_beam = mtf.Dimension("double_beam", beam_dim.size * 2)

    if use_tpu and layout is not None and mesh_shape is not None:
      # Do some partial top-k-ing first locally to avoid communication.
      # We reshape the logits from:
      #   [batch, beam, vocab] to
      #   [batch, beam, major_vocab, minor_vocab]
      # We first reduce (locally) across the minor_vocab dimension.  This makes
      # the thing we need to broadcast smaller.
      # This also enables our shortcut of only picking the top num_prefilter
      #   sequences per beam per major_vocab in the first pass.
      major_vocab_size = mtf.tensor_dim_to_mesh_dim_size(
          layout, mesh_shape, vocab_dim)
      major_vocab = mtf.Dimension(vocab_dim.name, major_vocab_size)
      minor_vocab = mtf.Dimension(
          "minor_vocab", vocab_dim.size // major_vocab_size)
      curr_scores = mtf.reshape(
          curr_scores, [batch_dim, beam_dim, major_vocab, minor_vocab])
      prefilter = mtf.Dimension("prefilter", num_prefilter or double_beam.size)
      # shape = [batch_dim, beam_dim, major_vocab, prefilter]
      top_scores, top_minor_vocab_ids = mtf.top_k(
          curr_scores, reduced_dim=minor_vocab, k_dim=prefilter)
      combined = mtf.Dimension(
          "combined", beam_dim.size * major_vocab.size * prefilter.size)
      top_scores = mtf.reshape(top_scores, [batch_dim, combined])
      top_minor_vocab_ids = mtf.reshape(
          top_minor_vocab_ids, [batch_dim, combined])
      # shpae = [batch_dim, double_beam]
      # ids are indices representing (beam, major_vocab, prefilter)
      top_scores, top_combined_ids = mtf.top_k(
          top_scores, reduced_dim=combined, k_dim=double_beam)
      top_minor_vocab_ids = mtf.gather(
          top_minor_vocab_ids, top_combined_ids, combined,
          output_shape=[batch_dim, double_beam])
      top_beam_index = top_combined_ids // (major_vocab.size * prefilter.size)
      top_combined_ids -= top_beam_index * (major_vocab.size * prefilter.size)
      top_major_vocab_ids = top_combined_ids // prefilter.size
      top_combined_ids -= top_major_vocab_ids * prefilter.size
      top_ids = top_major_vocab_ids * minor_vocab.size + top_minor_vocab_ids
    else:
      beam_and_vocab_dim = mtf.Dimension(
          "beam_and_vocab", beam_dim.size * vocab_dim.size)
      flat_shape = mtf.Shape([batch_dim, beam_and_vocab_dim])
      # Flatten out (beam_size, vocab_size) probs into a list of possibilities
      flat_curr_scores = mtf.reshape(
          curr_scores, flat_shape, name="flatten_scores")
      top_scores, top_ids = mtf.top_k(
          flat_curr_scores, reduced_dim=beam_and_vocab_dim, k_dim=double_beam)
      # Work out what beam the top probs are in.
      top_beam_index = top_ids // vocab_dim.size
      top_ids %= vocab_dim.size  # Unflatten the ids

    # Recovering the log probs because we will need to send them back
    top_log_probs = top_scores * length_penalty

    selector = mtf.one_hot(top_beam_index, beam_dim, dtype=tf.float32)

    def my_gather(tensor):
      return mtf.gather(
          tensor, top_beam_index, beam_dim,
          output_shape=mtf.Shape(
              [double_beam if d == beam_dim else d for d in tensor.shape.dims]))

    # Gather up the most probable 2*beams both for the ids and finished_in_alive
    # bools
    top_seq = my_gather(alive_seq)

    # Append the most probable alive
    top_seq += top_ids * mtf.one_hot(i, length_dim, dtype=tf.int32)
    top_finished = mtf.equal(top_ids, eos_id)

    return (
        top_seq, top_log_probs, top_scores, top_finished, new_states, selector)

  def inner_loop(i, alive_seq, alive_log_probs, finished_seq, finished_scores,
                 finished_flags, *states):
    """Inner beam search loop.

    There are three groups of tensors, alive, finished, and topk.
    The alive group contains information about the current alive sequences
    The topk group contains information about alive + topk current decoded words
    the finished group contains information about finished sentences, that is,
    the ones that have decoded to <EOS>. These are what we return.
    The general beam search algorithm is as follows:
    While we haven't terminated (pls look at termination condition)
      1. Grow the current alive to get beam*2 topk sequences
      2. Among the topk, keep the top beam_size ones that haven't reached EOS
      into alive
      3. Among the topk, keep the top beam_size ones have reached EOS into
      finished
    Repeat
    To make things simple with using fixed size tensors, we will end
    up inserting unfinished sequences into finished in the beginning. To stop
    that we add -ve INF to the score of the unfinished sequence so that when a
    true finished sequence does appear, it will have a higher score than all the
    unfinished ones.

    Args:
      i: loop index
      alive_seq: Topk sequences decoded so far [batch_size, beam_size, i+1]
      alive_log_probs: probabilities of the beams. [batch_size, beam_size]
      finished_seq: Current finished sequences.
        [batch_size, beam_size, i+1]
      finished_scores: scores for each of these sequences.
        [batch_size, beam_size]
      finished_flags: finished bools for each of these sequences.
        [batch_size, beam_size]
      *states: mtf Tensors

    Returns:
      Tuple of
        (Incremented loop index
         New alive sequences,
         Log probs of the alive sequences,
         New finished sequences,
         Scores of the new finished sequences,
         Flags indicating which sequence in finished as reached EOS,
         dict of final decoding states)
    """
    states = [mtf.replace_dimensions(
        state, batch_and_beam_dim, [batch_dim, beam_dim]) for state in states]
    # Each inner loop, we carry out three steps:
    # 1. Get the current topk items.
    # 2. Extract the ones that have finished and haven't finished
    # 3. Recompute the contents of finished based on scores.
    (top2k_seq, top2k_log_probs, top2k_scores, top2k_finished,
     new_states, first_selector) = grow_topk(
         i, alive_seq, alive_log_probs, states)
    with tf.variable_scope("grow_alive"):
      alive_seq, alive_log_probs, _, second_selector = grow_alive(
          top2k_seq, top2k_scores, top2k_log_probs, top2k_finished)
    with tf.variable_scope("grow_finished"):
      finished_seq, finished_scores, finished_flags, _ = grow_finished(
          finished_seq, finished_scores, finished_flags, top2k_seq,
          top2k_scores, top2k_finished)
    old_beam_dim = mtf.Dimension("old_beam", beam_dim.size)
    selector = mtf.einsum(
        [mtf.rename_dimension(first_selector, beam_dim.name, old_beam_dim.name),
         second_selector],
        output_shape=[batch_dim, old_beam_dim, beam_dim])
    gathered_states = []
    if use_tpu and layout is not None and mesh_shape is not None:
      # This hack combines the beam dimension with some of the batch dimension.
      # It makes gathering faster on TPU.
      #
      # Instead of multiplying by a [beam, beam] selector matrix, we instead
      # multiply by a [minor_batch*beam, minor_batch*beam] selector matrix.
      # This is theoretically more FLOPs, but it brings the matrix size closer
      # to the magic optimal value of 128.
      #
      # TODO(noam): file a bug with the XLA team to do this automatically
      major_batch_size = mtf.tensor_dim_to_mesh_dim_size(
          layout, mesh_shape, batch_dim)
      major_batch = mtf.Dimension(batch_dim.name, major_batch_size)
      minor_batch = mtf.Dimension(
          "minor_batch", batch_dim.size // major_batch.size)
      old_minor_batch = mtf.Dimension("old_minor_batch", minor_batch.size)
      old_combined = mtf.Dimension(
          "old_combined", minor_batch.size * beam_dim.size)
      combined = mtf.Dimension(
          "new_combined", old_combined.size)
      same_minor_batch = mtf.to_float(
          mtf.equal(mtf.range(mesh, old_minor_batch, tf.float32),
                    mtf.range(mesh, minor_batch, tf.float32)))
      selector = mtf.reshape(
          selector, [major_batch, minor_batch, old_beam_dim, beam_dim])
      selector = mtf.einsum(
          [selector, same_minor_batch],
          output_shape=[major_batch,
                        old_minor_batch, old_beam_dim,
                        minor_batch, beam_dim],
          reduced_dims=[])
      selector = mtf.reshape(selector, [major_batch, old_combined, combined])
      for state in new_states:
        s = mtf.replace_dimensions(
            state, [batch_dim, beam_dim], [major_batch, old_combined])
        s = mtf.einsum(
            [s, mtf.cast(selector, state.dtype)],
            reduced_dims=[old_combined],
            output_shape=mtf.replace_dimensions(
                state.shape, [batch_dim, beam_dim],
                [major_batch, combined]))
        gathered_states.append(mtf.replace_dimensions(
            s, [major_batch, combined], batch_and_beam_dim))
    else:
      for state in new_states:
        state = mtf.einsum(
            [mtf.rename_dimension(state, beam_dim.name, old_beam_dim.name),
             mtf.cast(selector, state.dtype)],
            reduced_dims=[old_beam_dim], output_shape=state.shape)
        state = mtf.replace_dimensions(
            state, [batch_dim, beam_dim], batch_and_beam_dim)
        gathered_states.append(state)

    return (i + 1, alive_seq, alive_log_probs, finished_seq, finished_scores,
            finished_flags) + tuple(gathered_states)

  def _is_finished(i, unused_alive_seq, alive_log_probs, unused_finished_seq,
                   finished_scores, finished_in_finished, *unused_states):
    """Checking termination condition.

    We terminate when we decoded up to decode_length or the lowest scoring item
    in finished has a greater score that the highest prob item in alive divided
    by the max length penalty

    Args:
      i: loop index
      alive_log_probs: probabilities of the beams. [batch_size, beam_size]
      finished_scores: scores for each of these sequences.
        [batch_size, beam_size]
      finished_in_finished: finished bools for each of these sequences.
        [batch_size, beam_size]

    Returns:
      Bool.
    """
    # TODO(noam): support a different decode length...
    # decode_length = mtf.constant(mesh, length_dim.size, dtype=tf.int32)

    # del alive_log_probs, finished_scores, finished_in_finished
    # return mtf.less(i, length_dim.size)
    if not stop_early:
      return mtf.less(i, decode_length)
    max_length_penalty = mtf.pow(
        ((5. + mtf.cast(decode_length, finished_scores.dtype)) / 6.), alpha)
    # The best possible score of the most likely alive sequence.
    lower_bound_alive_scores = mtf.gather(
        alive_log_probs, mtf.constant(mesh, 0, dtype=tf.int32),
        beam_dim) / max_length_penalty

    # Now to compute the lowest score of a finished sequence in finished
    # If the sequence isn't finished, we multiply it's score by 0. since
    # scores are all -ve, taking the min will give us the score of the lowest
    # finished item.
    lowest_score_of_finished_in_finished = mtf.reduce_min(
        finished_scores * mtf.cast(finished_in_finished, finished_scores.dtype),
        reduced_dim=beam_dim)

    # If none of the sequences have finished, then the min will be 0 and
    # we have to replace it by -ve INF if it is. The score of any seq in alive
    # will be much higher than -ve INF and the termination condition will not
    # be met.
    lowest_score_of_finished_in_finished += (
        (1. - mtf.cast(mtf.reduce_any(
            finished_in_finished, reduced_dim=beam_dim),
                       finished_scores.dtype)) * -INF)

    bound_is_met = mtf.reduce_all(
        mtf.greater(lowest_score_of_finished_in_finished,
                    lower_bound_alive_scores))
    return mtf.logical_and(
        mtf.less(i, decode_length), mtf.logical_not(bound_is_met))

  initial_step_num = mtf.constant(mesh, 0, dtype=tf.int32)
  states = [mtf.replace_dimensions(
      state, [batch_dim, beam_dim], batch_and_beam_dim) for state in states]
  while_loop_inputs = [
      initial_step_num, alive_seq, alive_log_probs, finished_seq,
      finished_scores, finished_flags] + states

  (_, alive_seq, alive_log_probs, finished_seq, finished_scores,
   finished_flags) = mtf.while_loop(
       _is_finished, inner_loop, while_loop_inputs,
       num_loop_vars=None if use_tpu else 6)[:6]

  # Accounting for corner case: It's possible that no sequence in alive for a
  # particular batch item ever reached EOS. In that case, we should just copy
  # the contents of alive for that batch item. tf.reduce_any(finished_flags, 1)
  # if 0, means that no sequence for that batch index had reached EOS. We need
  # to do the same for the scores as well.
  finished_seq = mtf.where(
      mtf.reduce_any(finished_flags, reduced_dim=beam_dim),
      finished_seq, alive_seq)
  finished_scores = mtf.where(
      mtf.reduce_any(finished_flags, reduced_dim=beam_dim),
      finished_scores, alive_log_probs)
  return finished_seq, finished_scores


@gin.configurable
def greedy_decode(logits_fn,
                  initial_ids,
                  temperature=0.0,
                  initial_states=None,
                  eos_id=EOS_ID,
                  forced_ids=None,
                  use_tpu=True):
  """Greedy decoding.

  Args:
    logits_fn: Interface to the model, to provide logits.
        Shoud take:
          step_num - mtf Scalar
          ids - mtf Tensor with shape [..., length]
          states - list of mtf.Tensor
        Should return:
          logits - [batch, vocab_size]
          new_states - list of mtf.Tensor
    initial_ids: mtf.Tensor with shape [..., length], containing zeros.
    temperature: a float between 0.0 (argmax) and 1.0 (random)
    initial_states: list of mtf.Tensor
    eos_id: ID for end of sentence.
    forced_ids: optional mtf.Tensor with shape [..., length]
    use_tpu: a boolean
  Returns:
    Tensor with shape [..., length]
  """
  length_dim = initial_ids.shape.dims[-1]
  mesh = initial_ids.mesh
  num_steps = mtf.constant(mesh, length_dim.size, dtype=tf.int32)
  def cond_fn(step_num, prev_ids, *unused_states):
    """Should we run another loop iteration."""
    overflow = mtf.equal(step_num, num_steps)
    has_eos = mtf.reduce_any(
        mtf.equal(prev_ids, eos_id), reduced_dim=length_dim)
    all_has_eos = mtf.reduce_all(has_eos)
    return mtf.logical_not(mtf.logical_or(overflow, all_has_eos))
  def body_fn(step_num, ids, *states):
    """Body function for greedy decoding.

    Args:
      step_num: a mtf.Tensor
      ids: a mtf.Tensor
      *states: additional mtf.Tensors
    Returns:
      new_step_num, new_ids, *new_states
    """
    logits, new_states = logits_fn(step_num, ids, states)
    vocab_dim = logits.shape.dims[-1]
    new_ids = mtf.sample_with_temperature(
        logits, vocab_dim, temperature)
    if forced_ids is not None:
      # force the new ids to equal the partial targets where specified
      # (positions where partial_targets contain nonzero values)
      forced = mtf.gather(forced_ids, step_num, length_dim)
      new_ids = forced + new_ids * mtf.to_int32(mtf.equal(forced, 0))
    ids += new_ids * mtf.one_hot(step_num, length_dim, dtype=tf.int32)
    new_step_num = step_num + 1
    return [new_step_num, ids] + new_states
  initial_step_num = mtf.constant(mesh, 0, dtype=tf.int32)
  while_loop_inputs = [initial_step_num, initial_ids] + initial_states
  final_step_num, mtf_samples = mtf.while_loop(
      cond_fn, body_fn, while_loop_inputs,
      num_loop_vars=None if use_tpu else 2)[:2]
  mtf_samples = mtf.Print(mtf_samples, [final_step_num], "output_length")
  return mtf_samples



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Auto MeshTensorflow subpackage."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from mesh_tensorflow.auto_mtf import api
from mesh_tensorflow.auto_mtf import graph_interface
from mesh_tensorflow.auto_mtf import layout_optimizer
from mesh_tensorflow.auto_mtf import memory_estimator
from mesh_tensorflow.auto_mtf import print_cp_model_solution
from mesh_tensorflow.auto_mtf import scheduler
from mesh_tensorflow.auto_mtf import valid_layouts
from mesh_tensorflow.auto_mtf.api import layout
from mesh_tensorflow.auto_mtf.api import layout_and_mesh_shape



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data augmentation lib for the Liver Tumor Segmentation (LiTS) dataset.

A set of data augmentation functions.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import tensorflow.compat.v1 as tf
import tensorflow_probability as tfp
from tensorflow.contrib import image as contrib_image


def _truncated_normal(mean, stddev):
  v = tf.random.normal(shape=[], mean=mean, stddev=stddev)
  v = tf.clip_by_value(v, -2 * stddev + mean, 2 * stddev + mean)
  return v


def _rand_noise(noise_mean, noise_dev, scale, shape):
  """Generate random noise given a particular scale and shape."""
  noise_shape = [x // scale for x in shape]
  noise_shape = [1 if x == 0 else x for x in noise_shape]
  noise = tf.random.normal(
      shape=noise_shape, mean=noise_mean, stddev=noise_dev)
  noise = tf.clip_by_value(
      noise, noise_mean - 2.0 * noise_dev, noise_mean + 2.0 * noise_dev)

  if scale != 1:
    noise = tf.image.resize_images(
        noise, [shape[0], shape[1]])
    noise = tf.transpose(noise, [0, 2, 1])
    noise = tf.image.resize_images(
        noise, [shape[0], shape[2]])
    noise = tf.transpose(noise, [0, 2, 1])
  return noise


def projective_transform(
    image, label, reso, image_translate_ratio, image_transform_ratio,
    sampled_2d_slices=False):
  """Apply projective transformation on image and label."""

  if image_translate_ratio < 0.000001 and (
      image_transform_ratio < 0.000001):
    return image, label

  def _projective_transform(data, proj_matrix, static_axis, interpolation):
    """Apply projective transformation."""
    if static_axis == 2:
      data = contrib_image.transform(data, proj_matrix, interpolation)
    elif static_axis == 1:
      data = tf.transpose(data, [0, 2, 1])
      data = contrib_image.transform(data, proj_matrix, interpolation)
      data = tf.transpose(data, [0, 2, 1])
    else:
      data = tf.transpose(data, [2, 1, 0])
      data = contrib_image.transform(data, proj_matrix, interpolation)
      data = tf.transpose(data, [2, 1, 0])
    return data

  for static_axis in [0, 1, 2]:
    if sampled_2d_slices and static_axis != 2:
      continue
    a0 = _truncated_normal(1.0, image_transform_ratio)
    a1 = _truncated_normal(0.0, image_transform_ratio)
    a2 = _truncated_normal(
        0.0, image_translate_ratio * reso)
    b0 = _truncated_normal(0.0, image_transform_ratio)
    b1 = _truncated_normal(1.0, image_transform_ratio)
    b2 = _truncated_normal(
        0.0, image_translate_ratio * reso)
    c0 = _truncated_normal(0.0, image_transform_ratio)
    c1 = _truncated_normal(0.0, image_transform_ratio)
    proj_matrix = [a0, a1, a2, b0, b1, b2, c0, c1]

    image = _projective_transform(image, proj_matrix, static_axis, 'BILINEAR')
    label = _projective_transform(label, proj_matrix, static_axis, 'NEAREST')

  return image, label


def maybe_add_noise(image, noise_shape, scale0, scale1,
                    image_noise_probability, image_noise_ratio):
  """Add noise at two scales."""

  if image_noise_probability < 0.000001 or (
      image_noise_ratio < 0.000001):
    return image

  noise_list = []
  for scale in [scale0, scale1]:
    rand_image_noise_ratio = tf.random.uniform(
        shape=[], minval=0.0, maxval=image_noise_ratio)
    noise_list.append(
        _rand_noise(0.0, rand_image_noise_ratio, scale, noise_shape))

  skip_noise = tf.greater(tf.random.uniform([]), image_noise_probability)
  image = tf.cond(skip_noise,
                  lambda: image, lambda: image + noise_list[0])
  image = tf.cond(skip_noise,
                  lambda: image, lambda: image + noise_list[1])

  return image


def _gen_rand_mask(ratio_mean, ratio_stddev, scale, shape, smoothness=0):
  """Generate a binary mask."""
  scale = max(scale, 1)

  ratio = tf.random.normal(
      shape=[], mean=ratio_mean, stddev=ratio_stddev)
  low_bound = tf.maximum(0.0, ratio_mean - 2 * ratio_stddev)
  up_bound = tf.minimum(1.0, ratio_mean + 2 * ratio_stddev)
  percentil_q = tf.cast(
      100.0 * tf.clip_by_value(ratio, low_bound, up_bound),
      tf.int32)

  pattern = _rand_noise(0.0, 1.0, scale, shape)
  if smoothness > 0:
    smoothness = int(smoothness) // 2 * 2 + 1
    pattern = tf.expand_dims(tf.expand_dims(pattern, 0), -1)
    pattern = tf.nn.conv3d(
        pattern, filter=tf.ones([smoothness, smoothness, smoothness, 1, 1]),
        strides=[1, 1, 1, 1, 1], padding='SAME', dilations=[1, 1, 1, 1, 1])
    pattern = tf.reduce_sum(pattern, 0)
    pattern = tf.reduce_sum(pattern, -1)

  thres = tfp.stats.percentile(pattern, q=percentil_q)
  rand_mask = tf.less(pattern, thres)

  return rand_mask


def maybe_gen_fake_data_based_on_real_data(
    image, label, reso, min_fake_lesion_ratio, gen_fake_probability):
  """Remove real lesion and synthesize lesion."""
  # TODO(lehou): Replace magic numbers with flag variables.
  gen_prob_indicator = tf.random_uniform(
      shape=[], minval=0.0, maxval=1.0, dtype=tf.float32)

  background_mask = tf.less(label, 0.5)
  lesion_mask = tf.greater(label, 1.5)
  liver_mask = tf.logical_not(tf.logical_or(background_mask, lesion_mask))

  liver_intensity = tf.boolean_mask(image, liver_mask)
  lesion_intensity = tf.boolean_mask(image, lesion_mask)

  intensity_diff = tf.reduce_mean(liver_intensity) - (
      tf.reduce_mean(lesion_intensity))
  intensity_diff *= 1.15
  intensity_diff = tf.cond(tf.is_nan(intensity_diff),
                           lambda: 0.0, lambda: intensity_diff)

  lesion_liver_ratio = 0.0
  lesion_liver_ratio += tf.random.normal(shape=[], mean=0.01, stddev=0.01)
  lesion_liver_ratio += tf.random.normal(shape=[], mean=0.0, stddev=0.05)
  lesion_liver_ratio = tf.clip_by_value(
      lesion_liver_ratio, min_fake_lesion_ratio, min_fake_lesion_ratio + 0.20)

  fake_lesion_mask = tf.logical_and(
      _gen_rand_mask(ratio_mean=lesion_liver_ratio, ratio_stddev=0.0,
                     scale=reso // 32, shape=label.shape,
                     smoothness=reso // 32),
      tf.logical_not(background_mask))
  liver_mask = tf.logical_not(tf.logical_or(background_mask, fake_lesion_mask))

  # Blur the masks
  lesion_mask_blur = tf.squeeze(tf.nn.conv3d(
      tf.expand_dims(tf.expand_dims(tf.cast(lesion_mask, tf.float32), -1), 0),
      filter=tf.ones([reso // 32] * 3 + [1, 1], tf.float32) / (reso // 32) ** 3,
      strides=[1, 1, 1, 1, 1],
      padding='SAME'))
  fake_lesion_mask_blur = tf.squeeze(tf.nn.conv3d(
      tf.expand_dims(tf.expand_dims(
          tf.cast(fake_lesion_mask, tf.float32), -1), 0),
      filter=tf.ones([reso // 32] * 3 + [1, 1], tf.float32) / (reso // 32) ** 3,
      strides=[1, 1, 1, 1, 1],
      padding='SAME'))

  # Remove real lesion and add fake lesion.
  # If the intensitify is too small (maybe no liver or lesion region labeled),
  # do not generate fake data.
  gen_prob_indicator = tf.cond(
      tf.greater(intensity_diff, 0.0001),
      lambda: gen_prob_indicator, lambda: 0.0)
  # pylint: disable=g-long-lambda
  image = tf.cond(
      tf.greater(gen_prob_indicator, 1 - gen_fake_probability),
      lambda: image + intensity_diff * lesion_mask_blur \
                    - intensity_diff * fake_lesion_mask_blur,
      lambda: image)
  label = tf.cond(
      tf.greater(gen_prob_indicator, 1 - gen_fake_probability),
      lambda: tf.cast(background_mask, tf.float32) * 0 + \
          tf.cast(liver_mask, tf.float32) * 1 + \
          tf.cast(fake_lesion_mask, tf.float32) * 2,
      lambda: label)
  # pylint: enable=g-long-lambda

  return image, label


def maybe_flip(image, label, flip_axis, flip_indicator=None):
  """Randomly flip the image."""
  if flip_indicator is None:
    flip_indicator = tf.random_uniform(shape=[])
  flip_or_not = tf.greater(flip_indicator, 0.5)

  def _maybe_flip(data):
    """Flip or not according to flip_or_not."""
    data = tf.cond(tf.logical_and(flip_or_not, tf.equal(flip_axis, 1)),
                   lambda: tf.transpose(data, [1, 0, 2]),
                   lambda: data)
    data = tf.cond(tf.logical_and(flip_or_not, tf.equal(flip_axis, 2)),
                   lambda: tf.transpose(data, [2, 1, 0]),
                   lambda: data)

    data = tf.cond(flip_or_not,
                   lambda: tf.image.flip_up_down(data),
                   lambda: data)

    data = tf.cond(tf.logical_and(flip_or_not, tf.equal(flip_axis, 1)),
                   lambda: tf.transpose(data, [1, 0, 2]),
                   lambda: data)
    data = tf.cond(tf.logical_and(flip_or_not, tf.equal(flip_axis, 2)),
                   lambda: tf.transpose(data, [2, 1, 0]),
                   lambda: data)
    return data

  return _maybe_flip(image), _maybe_flip(label)


def maybe_rot180(image, label, static_axis, rot180_k=None):
  """Randomly rotate the image 180 degrees."""
  if rot180_k is None:
    rot180_k = 2 * tf.random_uniform(
        shape=[], minval=0, maxval=2, dtype=tf.int32)
  rot_or_not = tf.not_equal(rot180_k, 0)

  def _maybe_rot180(data):
    """Rotate or not according to rot_or_not."""
    data = tf.cond(tf.logical_and(rot_or_not, tf.equal(static_axis, 0)),
                   lambda: tf.transpose(data, [2, 1, 0]),
                   lambda: data)
    data = tf.cond(tf.logical_and(rot_or_not, tf.equal(static_axis, 1)),
                   lambda: tf.transpose(data, [0, 2, 1]),
                   lambda: data)

    data = tf.cond(rot_or_not,
                   lambda: tf.image.rot90(data, k=rot180_k),
                   lambda: data)

    data = tf.cond(tf.logical_and(rot_or_not, tf.equal(static_axis, 0)),
                   lambda: tf.transpose(data, [2, 1, 0]),
                   lambda: data)
    data = tf.cond(tf.logical_and(rot_or_not, tf.equal(static_axis, 1)),
                   lambda: tf.transpose(data, [0, 2, 1]),
                   lambda: data)
    return data

  return _maybe_rot180(image), _maybe_rot180(label)


def intensity_shift(
    image, label, per_class_intensity_scale, per_class_intensity_shift):
  """Perturb intensity in lesion and non-lesion regions."""

  if per_class_intensity_scale < 0.000001 and (
      per_class_intensity_shift < 0.000001):
    return image

  # Randomly change (mostly increase) intensity of non-lesion region.
  per_class_noise = _truncated_normal(
      per_class_intensity_shift, per_class_intensity_scale)
  image = image + per_class_noise * (
      image * tf.cast(tf.greater(label, 1.5), tf.float32))

  # Randomly change (mostly decrease) intensity of lesion region.
  per_class_noise = _truncated_normal(
      -per_class_intensity_shift, per_class_intensity_scale)
  image = image + per_class_noise * (
      image * tf.cast(tf.less(label, 1.5), tf.float32))

  return image


def image_corruption(
    image, label, reso, image_corrupt_ratio_mean, image_corrupt_ratio_stddev):
  """Randomly drop non-lesion pixels."""

  if image_corrupt_ratio_mean < 0.000001 and (
      image_corrupt_ratio_stddev < 0.000001):
    return image

  # Corrupt non-lesion region according to keep_mask.
  keep_mask = _gen_rand_mask(
      1 - image_corrupt_ratio_mean,
      image_corrupt_ratio_stddev,
      reso // 3, image.shape)

  keep_mask = tf.logical_or(tf.greater(label, 1.5), keep_mask)
  image *= tf.cast(keep_mask, tf.float32)

  return image



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MNIST using Mesh TensorFlow and TF Estimator.

This is an illustration, not a good model.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import mesh_tensorflow as mtf
import mnist_dataset as dataset  # local file import
import tensorflow.compat.v1 as tf


tf.flags.DEFINE_string("data_dir", "/tmp/mnist_data",
                       "Path to directory containing the MNIST dataset")
tf.flags.DEFINE_string("model_dir", "/tmp/mnist_model", "Estimator model_dir")
tf.flags.DEFINE_integer("batch_size", 200,
                        "Mini-batch size for the training. Note that this "
                        "is the global batch size and not the per-shard batch.")
tf.flags.DEFINE_integer("hidden_size", 512, "Size of each hidden layer.")
tf.flags.DEFINE_integer("train_epochs", 40, "Total number of training epochs.")
tf.flags.DEFINE_integer("epochs_between_evals", 1,
                        "# of epochs between evaluations.")
tf.flags.DEFINE_integer("eval_steps", 0,
                        "Total number of evaluation steps. If `0`, evaluation "
                        "after training is skipped.")
tf.flags.DEFINE_string("mesh_shape", "b1:2;b2:2", "mesh shape")
tf.flags.DEFINE_string("layout", "row_blocks:b1;col_blocks:b2",
                       "layout rules")

FLAGS = tf.flags.FLAGS


def mnist_model(image, labels, mesh):
  """The model.

  Args:
    image: tf.Tensor with shape [batch, 28*28]
    labels: a tf.Tensor with shape [batch] and dtype tf.int32
    mesh: a mtf.Mesh

  Returns:
    logits: a mtf.Tensor with shape [batch, 10]
    loss: a mtf.Tensor with shape []
  """
  batch_dim = mtf.Dimension("batch", FLAGS.batch_size)
  row_blocks_dim = mtf.Dimension("row_blocks", 4)
  col_blocks_dim = mtf.Dimension("col_blocks", 4)
  rows_dim = mtf.Dimension("rows_size", 7)
  cols_dim = mtf.Dimension("cols_size", 7)

  classes_dim = mtf.Dimension("classes", 10)
  one_channel_dim = mtf.Dimension("one_channel", 1)

  x = mtf.import_tf_tensor(
      mesh, tf.reshape(image, [FLAGS.batch_size, 4, 7, 4, 7, 1]),
      mtf.Shape(
          [batch_dim, row_blocks_dim, rows_dim,
           col_blocks_dim, cols_dim, one_channel_dim]))
  x = mtf.transpose(x, [
      batch_dim, row_blocks_dim, col_blocks_dim,
      rows_dim, cols_dim, one_channel_dim])

  # add some convolutional layers to demonstrate that convolution works.
  filters1_dim = mtf.Dimension("filters1", 16)
  filters2_dim = mtf.Dimension("filters2", 16)
  f1 = mtf.relu(mtf.layers.conv2d_with_blocks(
      x, filters1_dim, filter_size=[9, 9], strides=[1, 1], padding="SAME",
      h_blocks_dim=row_blocks_dim, w_blocks_dim=col_blocks_dim, name="conv0"))
  f2 = mtf.relu(mtf.layers.conv2d_with_blocks(
      f1, filters2_dim, filter_size=[9, 9], strides=[1, 1], padding="SAME",
      h_blocks_dim=row_blocks_dim, w_blocks_dim=col_blocks_dim, name="conv1"))
  x = mtf.reduce_mean(f2, reduced_dim=filters2_dim)

  # add some fully-connected dense layers.
  hidden_dim1 = mtf.Dimension("hidden1", FLAGS.hidden_size)
  hidden_dim2 = mtf.Dimension("hidden2", FLAGS.hidden_size)

  h1 = mtf.layers.dense(
      x, hidden_dim1,
      reduced_dims=x.shape.dims[-4:],
      activation=mtf.relu, name="hidden1")
  h2 = mtf.layers.dense(
      h1, hidden_dim2,
      activation=mtf.relu, name="hidden2")
  logits = mtf.layers.dense(h2, classes_dim, name="logits")
  if labels is None:
    loss = None
  else:
    labels = mtf.import_tf_tensor(
        mesh, tf.reshape(labels, [FLAGS.batch_size]), mtf.Shape([batch_dim]))
    loss = mtf.layers.softmax_cross_entropy_with_logits(
        logits, mtf.one_hot(labels, classes_dim), classes_dim)
    loss = mtf.reduce_mean(loss)
  return logits, loss


def model_fn(features, labels, mode, params):
  """The model_fn argument for creating an Estimator."""
  tf.logging.info("features = %s labels = %s mode = %s params=%s" %
                  (features, labels, mode, params))
  global_step = tf.train.get_global_step()
  graph = mtf.Graph()
  mesh = mtf.Mesh(graph, "my_mesh")
  logits, loss = mnist_model(features, labels, mesh)
  mesh_shape = mtf.convert_to_shape(FLAGS.mesh_shape)
  layout_rules = mtf.convert_to_layout_rules(FLAGS.layout)
  mesh_size = mesh_shape.size
  mesh_devices = [""] * mesh_size
  mesh_impl = mtf.placement_mesh_impl.PlacementMeshImpl(
      mesh_shape, layout_rules, mesh_devices)

  if mode == tf.estimator.ModeKeys.TRAIN:
    var_grads = mtf.gradients(
        [loss], [v.outputs[0] for v in graph.trainable_variables])
    optimizer = mtf.optimize.AdafactorOptimizer()
    update_ops = optimizer.apply_grads(var_grads, graph.trainable_variables)

  lowering = mtf.Lowering(graph, {mesh: mesh_impl})
  restore_hook = mtf.MtfRestoreHook(lowering)

  tf_logits = lowering.export_to_tf_tensor(logits)
  if mode != tf.estimator.ModeKeys.PREDICT:
    tf_loss = lowering.export_to_tf_tensor(loss)
    tf.summary.scalar("loss", tf_loss)

  if mode == tf.estimator.ModeKeys.TRAIN:
    tf_update_ops = [lowering.lowered_operation(op) for op in update_ops]
    tf_update_ops.append(tf.assign_add(global_step, 1))
    train_op = tf.group(tf_update_ops)
    saver = tf.train.Saver(
        tf.global_variables(),
        sharded=True,
        max_to_keep=10,
        keep_checkpoint_every_n_hours=2,
        defer_build=False, save_relative_paths=True)
    tf.add_to_collection(tf.GraphKeys.SAVERS, saver)
    saver_listener = mtf.MtfCheckpointSaverListener(lowering)
    saver_hook = tf.train.CheckpointSaverHook(
        FLAGS.model_dir,
        save_steps=1000,
        saver=saver,
        listeners=[saver_listener])

    accuracy = tf.metrics.accuracy(
        labels=labels, predictions=tf.argmax(tf_logits, axis=1))

    # Name tensors to be logged with LoggingTensorHook.
    tf.identity(tf_loss, "cross_entropy")
    tf.identity(accuracy[1], name="train_accuracy")

    # Save accuracy scalar to Tensorboard output.
    tf.summary.scalar("train_accuracy", accuracy[1])

    # restore_hook must come before saver_hook
    return tf.estimator.EstimatorSpec(
        tf.estimator.ModeKeys.TRAIN, loss=tf_loss, train_op=train_op,
        training_chief_hooks=[restore_hook, saver_hook])

  if mode == tf.estimator.ModeKeys.PREDICT:
    predictions = {
        "classes": tf.argmax(tf_logits, axis=1),
        "probabilities": tf.nn.softmax(tf_logits),
    }
    return tf.estimator.EstimatorSpec(
        mode=tf.estimator.ModeKeys.PREDICT,
        predictions=predictions,
        prediction_hooks=[restore_hook],
        export_outputs={
            "classify": tf.estimator.export.PredictOutput(predictions)
        })
  if mode == tf.estimator.ModeKeys.EVAL:
    return tf.estimator.EstimatorSpec(
        mode=tf.estimator.ModeKeys.EVAL,
        loss=tf_loss,
        evaluation_hooks=[restore_hook],
        eval_metric_ops={
            "accuracy":
            tf.metrics.accuracy(
                labels=labels, predictions=tf.argmax(tf_logits, axis=1)),
        })


def run_mnist():
  """Run MNIST training and eval loop."""
  mnist_classifier = tf.estimator.Estimator(
      model_fn=model_fn,
      model_dir=FLAGS.model_dir)

  # Set up training and evaluation input functions.
  def train_input_fn():
    """Prepare data for training."""

    # When choosing shuffle buffer sizes, larger sizes result in better
    # randomness, while smaller sizes use less memory. MNIST is a small
    # enough dataset that we can easily shuffle the full epoch.
    ds = dataset.train(FLAGS.data_dir)
    ds_batched = ds.cache().shuffle(buffer_size=50000).batch(FLAGS.batch_size)

    # Iterate through the dataset a set number (`epochs_between_evals`) of times
    # during each training session.
    ds = ds_batched.repeat(FLAGS.epochs_between_evals)
    return ds

  def eval_input_fn():
    return dataset.test(FLAGS.data_dir).batch(
        FLAGS.batch_size).make_one_shot_iterator().get_next()

  # Train and evaluate model.
  for _ in range(FLAGS.train_epochs // FLAGS.epochs_between_evals):
    mnist_classifier.train(input_fn=train_input_fn, hooks=None)
    eval_results = mnist_classifier.evaluate(input_fn=eval_input_fn)
    print("\nEvaluation results:\n\t%s\n" % eval_results)


def main(_):
  run_mnist()


if __name__ == "__main__":
  tf.disable_v2_behavior()
  tf.logging.set_verbosity(tf.logging.INFO)
  tf.app.run()



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Layers for the Transformer."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import math
import gin

import mesh_tensorflow as mtf
from mesh_tensorflow import layers
from mesh_tensorflow.transformer import attention
from mesh_tensorflow.transformer import transformer

import tensorflow.compat.v1 as tf


@gin.configurable
class DenseReluDense(transformer.TransformerLayer):
  """Two dense layers with ReLU or other activation on hidden layer."""

  def __init__(self, hidden_size=4096, dropout_rate=0.0, activation="relu"):
    """Create a DenseReluDense.

    Args:
      hidden_size: an integer - size of the hidden layer
      dropout_rate: a floating-point number
      activation: an activation function or a list of activation functions.
        see documentation for mtf.layers.dense_product()
    """
    self.hidden_size = hidden_size
    self.dropout_rate = dropout_rate
    self.activation = activation

  def call(self, context, x, losses=None):
    """Call the layer."""
    io_channels = x.shape.dims[-1]
    hidden_channels = mtf.Dimension("d_ff", self.hidden_size)
    h = mtf.layers.dense_product(x,
                                 reduced_dims=x.shape.dims[-1:],
                                 new_dims=hidden_channels,
                                 activation_functions=self.activation,
                                 use_bias=False,
                                 variable_dtype=context.variable_dtype,
                                 name="wi",
                                 expert_dims=context.model.ensemble_dims)
    if context.train and self.dropout_rate != 0.0:
      h = mtf.dropout(h, 1.0 - self.dropout_rate,
                      noise_shape=h.shape - context.length_dim)
    return mtf.layers.dense(h, io_channels, use_bias=False, activation=None,
                            variable_dtype=context.variable_dtype,
                            reduced_dims=h.shape.dims[-1:],
                            name="wo",
                            expert_dims=context.model.ensemble_dims)


def attention_params(context,
                     kv_dim,
                     num_heads,
                     num_memory_heads=0,
                     shared_kv=False):
  """Attention Parameters for Transformer Layers.

  The num_heads argument indicates the number of read-heads.

  For the familiar behavior described in "Attention Is All You Need", set
  num_memory_heads=0.

  If num_memory_heads==1, then there is only a single write-head, and multiple
  read-heads.  This leads to faster incremental decoding, since the
  recurrent state is smaller

  If num_memory_heads > 1, then num_memory_heads indicates the number of
  write-heads.  A fraction of the read-heads read each write-head.
  num_memory_heads must divide num_heads. This behavior has not yet been tested.

  Args:
    context: a transformer.Context
    kv_dim: a dimension (for key and value channels)
    num_heads: an integer
    num_memory_heads: an optional integer
    shared_kv: a boolean
  Returns:
    an attention.AttentionParams object
  """
  if num_heads == 1:
    query_heads_dims = None
    memory_heads_dims = None
  elif num_memory_heads == 0:
    query_heads_dims = [mtf.Dimension("heads", num_heads)]
    memory_heads_dims = query_heads_dims
  elif num_memory_heads == 1:
    query_heads_dims = [mtf.Dimension("heads", num_heads)]
    memory_heads_dims = None
  else:
    if num_heads % num_memory_heads != 0:
      raise ValueError("num_memory_heads must divide num_heads")
    memory_heads_dims = [mtf.Dimension("heads", num_memory_heads)]
    query_heads_dims = memory_heads_dims + [
        mtf.Dimension("query_heads", num_heads // num_memory_heads)]
  return attention.AttentionParams(
      context.mesh,
      query_input_dim=context.model.model_dim,
      memory_input_dim=context.model.model_dim,
      output_dim=context.model.model_dim,
      key_dim=kv_dim,
      value_dim=kv_dim,
      query_heads_dims=query_heads_dims,
      memory_heads_dims=memory_heads_dims,
      variable_dtype=context.variable_dtype,
      shared_kv=shared_kv,
      ensemble_dim=context.model.ensemble_dim)


@gin.configurable
class SelfAttention(transformer.TransformerLayer):
  """Multi-head self-attention layer."""

  def __init__(self,
               num_heads=8,
               num_memory_heads=0,
               key_value_size=128,
               shared_kv=False,
               dropout_rate=0.0,
               attention_kwargs=None,
               relative_attention_type=None,
               relative_attention_num_buckets=32,
               attention_func=None):
    """Create a SelfAttention Layer.

    Args:
      num_heads: an integer
      num_memory_heads: an optional integer
      key_value_size: an integer
      shared_kv: a boolean
      dropout_rate: a float
      attention_kwargs: a dictionary of kwargs for attention.attention
      relative_attention_type: an optional string - one of
        (None, "bias", "bias_shared", "contextual")
      relative_attention_num_buckets: an integer
      attention_func: attention function: None/'hybrid'.
    """
    self.num_heads = num_heads
    self.num_memory_heads = num_memory_heads
    self.key_value_size = key_value_size
    self.shared_kv = shared_kv
    self.dropout_rate = dropout_rate
    self.attention_kwargs = attention_kwargs or {}
    self.relative_attention_type = relative_attention_type
    self.relative_attention_num_buckets = relative_attention_num_buckets
    self.attention_func = attention_func

  def attention_kwargs_from_context(self, context):
    kwargs = copy.copy(self.attention_kwargs)
    kwargs["dropout_rate"] = self.dropout_rate if context.train else 0.0
    if "dropout_broadcast_dims" not in kwargs:
      kwargs["dropout_broadcast_dims"] = [context.length_dim]
    return kwargs

  def make_params(self, context):
    return attention_params(context=context,
                            kv_dim=self.kv_dim,
                            num_heads=self.num_heads,
                            num_memory_heads=self.num_memory_heads,
                            shared_kv=self.shared_kv)

  def call(self, context, x, losses=None):
    """Call the layer."""
    params = self.make_params(context)
    q = params.compute_q(x)
    memory_length = self.memory_length(context)
    if context.mode == "incremental":
      m = x
    else:
      m = mtf.replace_dimensions(x, context.length_dim, memory_length)
    if self.shared_kv:
      kv = params.compute_kv(m)
    else:
      k = params.compute_k(m)
      v = params.compute_v(m)
    if context.mode == "incremental":
      one_hot = mtf.one_hot(
          context.position, memory_length, dtype=context.activation_dtype)
      inv_one_hot = 1.0 - one_hot
      if self.shared_kv:
        old_kv = context.get_states(1)
        kv = old_kv * inv_one_hot + kv * one_hot
      else:
        old_k, old_v = context.get_states(2)
        k = old_k * inv_one_hot + k * one_hot
        v = old_v * inv_one_hot + v * one_hot
      memory_position = mtf.range(context.mesh, memory_length, tf.int32)
    else:
      memory_position = self.rename_length_to_memory_length(
          context.position, context)
    if context.mode == "incremental" or context.mode == "first_part":
      context.record_new_states([kv] if self.shared_kv else [k, v])
    if self.shared_kv:
      k = kv
      v = kv
    if self.attention_func == "hybrid":
      o = attention.hybrid_attention(
          q, k, v, context, memory_length, self.kv_dim, self.kv_dim,
          self.compute_bias(context, memory_position, x,
                            params.query_heads_dims, q),
          **self.attention_kwargs_from_context(context))
    else:
      o = attention.attention(
          q, k, v, memory_length, self.kv_dim, self.kv_dim,
          self.compute_bias(context, memory_position, x,
                            params.query_heads_dims, q),
          **self.attention_kwargs_from_context(context))
    return params.compute_output(o, output_shape=x.shape)

  def compute_bias(self, context, memory_position, x, heads_dims, q):
    """Compute attention bias.

    Args:
      context: a transformer.Context
      memory_position: an int32 tensor containing memory_length dimension.
      x: a Tensor - the query antecedent - required for relative attention
      heads_dims: a list of dimensions
      q: a Tensor - the queries - required for contextual relative attention
    Returns:
      a Tensor or None
    """
    min_relative_position = self.min_relative_position(context)
    max_relative_position = self.max_relative_position(context)
    # we can often cache the result of this function between similar layers
    can_cache = (
        self.relative_attention_type is None or
        self.relative_attention_type == "bias_shared")
    if can_cache:
      cache_key = ("self_attention_mask",
                   min_relative_position,
                   max_relative_position,
                   self.relative_attention_type,
                   tuple(heads_dims))
      if cache_key in context.cache:
        return context.cache[cache_key]
    biases = []
    relative_position = memory_position - context.position
    if min_relative_position is not None:
      visible = mtf.greater_equal(relative_position, min_relative_position)
      biases.append(attention.visibility_mask_to_attention_bias(
          visible, context.activation_dtype))
    if max_relative_position is not None:
      visible = mtf.less_equal(relative_position, max_relative_position)
      biases.append(attention.visibility_mask_to_attention_bias(
          visible, context.activation_dtype))
    if context.read_priority is not None:
      visible = mtf.greater_equal(
          context.read_priority,
          mtf.layers.rename_length_to_memory_length(context.write_priority))
      biases.append(attention.visibility_mask_to_attention_bias(
          visible, context.activation_dtype))

    sequence_id = None
    # Subsequence id should only be set if we are in the decoder and have
    # multiple targets per input. This will allow each sub-target to only attend
    # to itself.
    if isinstance(context.subsequence_id, mtf.Tensor):
      sequence_id = context.subsequence_id
    elif isinstance(context.sequence_id, mtf.Tensor):
      sequence_id = context.sequence_id
    if (sequence_id is not None and context.length_dim in sequence_id.shape):
      visible = mtf.equal(
          sequence_id,
          self.rename_length_to_memory_length(sequence_id, context))
      biases.append(attention.visibility_mask_to_attention_bias(
          visible, context.activation_dtype))
    if self.relative_attention_type is not None:
      buckets_dim = mtf.Dimension(
          "buckets", self.relative_attention_num_buckets)
      bidirectional = not context.model.fully_autoregressive
      rp_bucket = _relative_position_bucket(
          relative_position,
          bidirectional=bidirectional,
          num_buckets=buckets_dim.size)
      if (self.relative_attention_type == "bias" or
          self.relative_attention_type == "bias_shared"):
        bias_shape = context.model.ensemble_dims + heads_dims + [buckets_dim]
        values = mtf.get_variable(
            context.mesh, "relative_attention_bias",
            bias_shape, dtype=context.variable_dtype)
      elif self.relative_attention_type == "contextual":
        values = layers.dense(
            q, reduced_dims=[self.kv_dim],
            new_dims=[buckets_dim],
            variable_dtype=context.variable_dtype,
            name="relative_attention_ak",
            use_bias=False,
            expert_dims=context.model.ensemble_dims + heads_dims)
      else:
        raise ValueError("unrecognized relative_attention_type \"%s\"" %
                         self.relative_attention_type)
      biases.append(mtf.gather(values, rp_bucket, buckets_dim))
    ret = mtf.add_n(biases) if biases else None
    if can_cache:
      context.cache[cache_key] = ret
    return ret

  @property
  def kv_dim(self):
    return mtf.Dimension("d_kv", self.key_value_size)

  def memory_length(self, context):
    return mtf.Dimension("memory_length", context.length_dim.size)

  def rename_length_to_memory_length(self, x, context):
    return mtf.replace_dimensions(
        x, context.length_dim, self.memory_length(context))

  def min_relative_position(self, context):
    return None

  def max_relative_position(self, context):
    return None


@gin.configurable
class EncDecAttention(SelfAttention):
  """Multi-head attention over encoder output."""

  def _get_memory_antecedent(self, context):
    return context.encoder_output

  def call(self, context, x, losses=None):
    """Call the layer."""
    memory_antecedent = self._get_memory_antecedent(context)
    memory_input_dim = memory_antecedent.shape[-1]
    if memory_input_dim != context.model.model_dim:
      raise NotImplementedError(
          "TODO(noam): support different model_dim in encoder and decoder.")
    params = self.make_params(context)
    q = params.compute_q(x)
    if context.mode == "incremental":
      k, v, memory_length = context.get_constant_state()
    else:
      m = memory_antecedent
      if self.shared_kv:
        kv = params.compute_kv(m)
        k = kv
        v = kv
      else:
        k = params.compute_k(m)
        v = params.compute_v(m)
      memory_length, = [d for d in m.shape.dims if d.name == "memory_length"]
      if context.mode == "first_part":
        context.record_constant_state((k, v, memory_length))
    if context.encoder_sequence_id and context.sequence_id:
      visible = mtf.equal(context.sequence_id, context.encoder_sequence_id)
      bias = attention.visibility_mask_to_attention_bias(
          visible, context.activation_dtype)
    else:
      bias = None
    o = attention.attention(
        q, k, v,
        memory_length,
        self.kv_dim,
        self.kv_dim,
        bias,
        **self.attention_kwargs_from_context(context))
    return params.compute_output(o, output_shape=x.shape)


@gin.configurable
class TransparentEncDecAttention(EncDecAttention):
  """Transparent multi-head attention over encoder output."""

  def __init__(self,
               layers_per_encoder_module=gin.REQUIRED,
               layers_per_decoder_module=gin.REQUIRED,
               encoder_num_modules=gin.REQUIRED,
               decoder_num_modules=gin.REQUIRED,
               dropout_rate=0.0,
               **kwargs):
    """Create a transparent attention EncDec Layer.

    Args:
      layers_per_encoder_module: positive integer telling how many layer are in
        each repeated module in the encoder
      layers_per_decoder_module: positive integer telling how many layer are in
        each repeated module in the decoder
      encoder_num_modules: positive integer of how many repeated modules there
        are in the encoder
      decoder_num_modules: positive integer of how many repeated modules there
        are in the decoder
      dropout_rate: positive float, the dropout rate for the matrix relating
        encoder outputs to decoder inputs
      **kwargs: additional constructor params
    """
    super(TransparentEncDecAttention, self).__init__(**kwargs)
    self.layers_per_encoder_module = layers_per_encoder_module
    self.layers_per_decoder_module = layers_per_decoder_module
    self.encoder_num_modules = encoder_num_modules
    self.decoder_num_modules = decoder_num_modules
    self.dropout_rate = dropout_rate

  def _get_memory_antecedent(self, context):
    decoder_module_index = context.layer_index // self.layers_per_decoder_module
    decoder_inputs = self._get_decoder_inputs(context)
    return decoder_inputs[decoder_module_index]

  def _get_decoder_inputs(self, context):
    """Computes the inputs to the decoder when using transparent attention.

    We must cache on the context in order to ensure that we are not replicating
    variables when the layer's call function is called in different tf variable
    scopes.

    Args:
      context: a Context

    Returns:
      a list containing `self.num_decoder_modules` of tensors with shape
        [<batch_dims>, length_dim, output_vocab_dim]
    """
    if hasattr(context, "decoder_layers_per_module"):
      return context.decoder_layers_per_module

    encoder_layer_outputs = [
        mtf.layers.rename_length_to_memory_length(output)
        for output in context.encoder_layer_outputs
    ]

    layers_per_module = self.layers_per_encoder_module
    encoder_module_outputs_dim = mtf.Dimension(
        "encoder_module_outputs", size=self.encoder_num_modules + 1)
    decoder_module_inputs_dim = mtf.Dimension(
        "decoder_module_inputs", size=self.decoder_num_modules)
    encoder_module_outputs = mtf.stack(
        [encoder_layer_outputs[0]] +
        encoder_layer_outputs[layers_per_module::layers_per_module],
        dim_name="encoder_module_outputs")
    w = mtf.get_variable(
        context.mesh,
        "w",
        mtf.Shape([encoder_module_outputs_dim, decoder_module_inputs_dim]),
        initializer=tf.random_normal_initializer(
            stddev=(encoder_module_outputs_dim.size *
                    decoder_module_inputs_dim.size)**-0.5),
        dtype=context.variable_dtype)
    if context.train and self.dropout_rate != 0.0:
      w = mtf.dropout(w, 1.0 - self.dropout_rate)
    s = mtf.softmax(w, reduced_dim=encoder_module_outputs_dim)
    z = mtf.einsum([s, encoder_module_outputs],
                   reduced_dims=[encoder_module_outputs_dim])
    input_per_decoder = mtf.split(
        z,
        split_dim=decoder_module_inputs_dim,
        num_or_size_splits=decoder_module_inputs_dim.size)
    context.decoder_layers_per_module = [
        mtf.reshape(inpt, z.shape.dims[1:]) for inpt in input_per_decoder
    ]
    return context.decoder_layers_per_module


@gin.configurable
class LocalSelfAttention(SelfAttention):
  """Multi-head local self-attention layer."""

  def __init__(self,
               radius=128,
               num_heads=8,
               num_memory_heads=0,
               key_value_size=128,
               shared_kv=False,
               dropout_rate=0.0,
               attention_kwargs=None,):
    super(LocalSelfAttention, self).__init__(
        num_heads,
        num_memory_heads,
        key_value_size,
        shared_kv,
        dropout_rate,
        attention_kwargs)
    self.radius = radius

  def call(self, context, x, losses=None):
    """Call the layer."""
    params = self.make_params(context)
    q = params.compute_q(x)
    if self.shared_kv:
      kv = params.compute_kv(x)
      k = kv
      v = kv
    else:
      k = params.compute_k(x)
      v = params.compute_v(x)
    if context.mode == "incremental":
      if self.shared_kv:
        prev_kv, = context.get_states(1)
      else:
        prev_k, prev_v = context.get_states(2)
      current_position = mtf.equal(
          mtf.range(context.mesh, self.window_dim, dtype=tf.int32),
          mtf.mod(context.position, self.radius))
      if self.shared_kv:
        kv = mtf.where(current_position, kv, prev_kv,
                       output_shape=prev_kv.shape)
        k = kv
        v = kv
        context.record_new_states([kv])
      else:
        k = mtf.where(current_position, params.compute_k(x), prev_k,
                      output_shape=prev_k.shape)
        v = mtf.where(current_position, params.compute_v(x), prev_v,
                      output_shape=prev_v.shape)
        context.record_new_states([k, v])
      window_pos = mtf.range(context.mesh, self.window_dim, tf.int32)
      visible = mtf.greater_equal(context.position, window_pos)
      bias = attention.visibility_mask_to_attention_bias(
          visible, context.activation_dtype)
      o = attention.attention(
          q,
          k,
          v,
          self.window_dim,
          self.kv_dim,
          self.kv_dim,
          bias,
          **self.attention_kwargs_from_context(context))
    elif context.length_dim.size <= max(256, self.radius * 4):
      # nothing fancy - just do full attention and mask
      memory_length = self.rename_length_to_memory_length(
          context.position, context)
      o = attention.attention(
          q, self.rename_length_to_memory_length(k, context),
          self.rename_length_to_memory_length(v, context),
          self.memory_length(context), self.kv_dim, self.kv_dim,
          self.compute_bias(context, memory_length, x, params.query_heads_dims,
                            q), **self.attention_kwargs_from_context(context))
    else:
      # fancy local attention algorithm
      o = attention.local_attention_1d(
          q=q,
          k=k,
          v=None if self.shared_kv else v,
          length_dim=context.length_dim,
          key_dim=self.kv_dim,
          value_dim=self.kv_dim,
          length_dim_num_splits=1,  # TODO(noam): look at the layout
          autoregressive=context.model.fully_autoregressive,
          radius=self.radius,
          sequence_id=context.sequence_id,
          write_priority=context.write_priority,
          read_priority=context.read_priority,
          attention_kwargs=self.attention_kwargs_from_context(context))
    if context.mode == "first_part":
      window_pos = mtf.range(context.mesh, self.window_dim, tf.int32)
      pos = mtf.range(context.mesh, context.length_dim, tf.int32)
      select_recent = mtf.cast(
          mtf.equal(mtf.mod(pos, self.radius), window_pos), x.dtype)
      select_recent *= mtf.cast(
          mtf.less(pos, context.initial_position), x.dtype)
      select_recent *= mtf.cast(
          mtf.greater_equal(
              pos, context.initial_position - self.radius), x.dtype)
      state_shape = (k.shape - [context.length_dim, self.kv_dim]
                     + [self.window_dim, self.kv_dim])
      k_state = mtf.einsum(
          [k, select_recent], output_shape=state_shape,
          reduced_dims=[context.length_dim])
      context.new_states.append(k_state)
      if not self.shared_kv:
        v_state = mtf.einsum(
            [v, select_recent], output_shape=state_shape,
            reduced_dims=[context.length_dim])
        context.new_states.append(v_state)
    return params.compute_output(o, output_shape=x.shape)

  def min_relative_position(self, context):
    return 1 - self.radius

  def max_relative_position(self, context):
    return None if context.model.fully_autoregressive else self.radius

  @property
  def window_dim(self):
    return mtf.Dimension("window", self.radius)


def _relative_position_bucket(relative_position,
                              bidirectional=True,
                              num_buckets=32,
                              max_distance=128):
  """Translate relative position to a bucket number for relative attention.

  The relative position is defined as memory_position - query_position, i.e.
  the distance in tokens from the attending position to the attended-to
  position.  If bidirectional=False, then positive relative positions are
  invalid.

  We use smaller buckets for small absolute relative_position and larger buckets
  for larger absolute relative_positions.  All relative positions >=max_distance
  map to the same bucket.  All relative positions <=-max_distance map to the
  same bucket.  This should allow for more graceful generalization to longer
  sequences than the model has been trained on.

  Args:
    relative_position: an int32 Tensor
    bidirectional: a boolean - whether the attention is bidirectional
    num_buckets: an integer
    max_distance: an integer
  Returns:
    a Tensor with the same shape as relative_position, containing int32
      values in the range [0, num_buckets)
  """
  ret = 0
  n = -relative_position
  if bidirectional:
    num_buckets //= 2
    ret += mtf.to_int32(mtf.less(n, 0)) * num_buckets
    n = mtf.abs(n)
  else:
    n = mtf.maximum(n, 0)
  # now n is in the range [0, inf)
  max_exact = num_buckets // 2
  is_small = mtf.less(n, max_exact)
  val_if_large = max_exact + mtf.to_int32(
      mtf.log(mtf.to_float(n) / max_exact)
      / math.log(max_distance / max_exact) * (num_buckets - max_exact))
  val_if_large = mtf.minimum(val_if_large, num_buckets - 1)
  ret += mtf.where(is_small, n, val_if_large)
  return ret





# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mesh TensorFlow."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from mesh_tensorflow.ops import *  # pylint: disable=wildcard-import
from mesh_tensorflow.ops import mtf_abs as abs  # pylint: disable=redefined-builtin,unused-import
from mesh_tensorflow.ops import mtf_pow as pow  # pylint: disable=redefined-builtin,unused-import
from mesh_tensorflow.ops import mtf_range as range  # pylint: disable=redefined-builtin,unused-import
from mesh_tensorflow.ops import mtf_slice as slice  # pylint: disable=redefined-builtin,unused-import



# TODO(trandustin): Seal module.
# from tensorflow.python.util.all_util import remove_undocumented  # pylint: disable=line-too-long
#
# _allowed_symbols = None
#
# remove_undocumented(__name__, _allowed_symbols)



# coding=utf-8
# Copyright 2020 The Mesh TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Learning rate schedules for training the transformer.

All learning rate schedule functions must take the scalar named argument `step`
and the numeric argument `total_train_steps`. They must output a tf.Scalar which
is the learning rate for the step.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

import gin
import tensorflow.compat.v1 as tf


@gin.configurable
def constant_learning_rate(step, total_train_steps, learning_rate=gin.REQUIRED):
  """Learning rate independent of step.

  Args:
    step: a tf.Scalar
    total_train_steps: a number
    learning_rate: a number or tf.Scalar

  Returns:
    a tf.Scalar, the learning rate for the step.
  """
  del step, total_train_steps
  return tf.cast(learning_rate, tf.float32)


@gin.configurable
def linear_decay_learning_rate(step,
                               total_train_steps,
                               initial_lr=0.1,
                               offset=0):
  """Linearly decay the learning rate to 0.

  Args:
    step: a tf.scalar representing the step we want the learning rate for.
    total_train_steps: a number, the total number of training steps.
    initial_lr: initial learning rate. Decays from here.
    offset: a number used for finetuning. Starts the learning-rate decay
      schedule from this step forwards.

  Returns:
    a tf.Scalar, the learning rate for the step.
  """
  offset = tf.cast(offset, tf.float32)
  step = tf.cast(step, tf.float32)

  return initial_lr * tf.minimum(1.0, (total_train_steps - step) /
                                 (total_train_steps - offset))


@gin.configurable
def learning_rate_schedule_noam(step,
                                total_train_steps,
                                warmup_steps=10000,
                                linear_decay_fraction=0.1,
                                multiplier=1.0,
                                offset=0):
  """Noam's favorite learning-rate schedule.

  (rsqrt(max(step_num, warmup_steps))
   * multiplier
   * min(1.0, (train_steps-step_num)/(train_steps*linear_decay_fraction)))

  Args:
    step: a tf.scalar representing the step we want the learning rate for.
    total_train_steps: a number, the total number of training steps.
    warmup_steps: a number
    linear_decay_fraction: a number
    multiplier: a number
    offset: a number used for finetuning. Starts the learning-rate decay
      schedule from this step forwards. Prior to this step, the learning rate is
      the same as if it were a warmup step.

  Returns:
    a tf.Scalar, the learning rate for the step.
  """
  train_steps = float(total_train_steps)
  step_num = tf.cast(step, tf.float32) - offset
  learning_rate = tf.math.rsqrt(tf.maximum(step_num, warmup_steps))
  learning_rate *= multiplier
  if linear_decay_fraction > 0:
    learning_rate *= tf.minimum(1.0, (train_steps - step_num) /
                                (train_steps * linear_decay_fraction))
  return learning_rate


@gin.configurable
def slanted_triangular(step,
                       total_train_steps,
                       cut_fraction=0.1,
                       ratio=32,
                       max_learning_rate=0.01,
                       start_step=0):
  """Triangular learning rate with short increase and long decay.

  Taken from "Universal Language Model Fine-tuning for Text Classification",
  see https://arxiv.org/abs/1801.06146. Default parameters are those specified
  in the paper.

  Args:
    step: a tf.scalar representing the step we want the learning rate for.
    total_train_steps: a number, the total number of training steps.
    cut_fraction: a number between 0 and 1, fraction of iterations for which we
      are increasing the learning rate.
    ratio: a number greater than 1, the ratio from the smallest learning rate to
      the max learning rate.
    max_learning_rate: a number, the highest learning rate reached during
      training.
    start_step: a number, the step training starts at. Useful when fine-tuning
      from a checkpoint that hasn't had its global step reset.

  Returns:
    a tf.Scalar, the learning rate for the step.
  """
  train_steps = float(total_train_steps)
  start_step = float(start_step)
  step_num = tf.cast(step, tf.float32) - start_step
  cut = math.floor(train_steps * cut_fraction)
  p = tf.cond(
      step_num < cut,
      lambda: step_num / cut,
      lambda: 1 - (step_num - cut) / (cut * (1 / cut_fraction - 1)),
  )
  return max_learning_rate * (1 + p * (ratio - 1)) / ratio
